"""
This module contains all of the qmmm namelist variables for the
amber programs and automatically loads those dictionaries with the
default values found in that program (sander only).
"""
from typing import Dict

from .typing import InputDataType
class qmmm:

    def __init__(self):
        self.sander: Dict[str, InputDataType] = {
            'qmcut' : -1, 'iqmatoms' : '', 'qmmask' : '', 'qmgb' : 2,
            'qm_theory' : '', 'qmcharge' : 0, 'qmqmdx' : 1,
            'verbosity' : 0, 'tight_p_conv' : 0, 'scfconv' : 1.0e-8,
            'errconv' : 1e-1, 'ndiis_matrices' :6, 'ndiis_attempts' : 0,
            'printcharges' : 0, 'printdipole' : 0, 'peptide_corr' : 0,
            'itrmax' : 1000, 'qmshake' : 1, 'qmqm_erep_incore' : 1,
            'qmmmrij_incore' : 1, 'lnk_dis' : 1.09, 'lnk_atomic_no' : 1,
            'lnk_method' : 1, 'spin' : 1, 'pseudo_diag' : 1,
            'pseudo_diag_criteria' : 0.05, 'qm_ewald' : 1, 'qm_pme' : 1,
            'kmaxqx' : 5, 'kmaxqy' : 5, 'kmaxqz' : 5, 'ksqmaxq' : 27,
            'writepdb' : 0, 'qmmm_int' : 1, 'adjust_q' : 2,
            'diag_routine' : 1, 'density_predict' : 0,
            'fock_predict' : 0, 'fockp_d1' : 2.4, 'fockp_d2' : -1.2,
            'fockp_d3' : -0.8, 'fockp_d4' : 0.6, 'idc' : 0, 'divpb' : 0,
            'dftb_maxiter' : 70, 'dftb_disper' : 0,
            'dftb_3rd_order' : 'NONE', 'dftb_chg' : 0,
            'dftb_telec' : 0, 'dftb_telec_step' : 0,
            'qmmm_omp_max_threads' : 1, 'chg_lambda' : 1,
            'nearest_qm_solvent' : 0, 'nearest_qm_solvent_fq' : 1,
            'nearest_qm_solvent_resname' : 'WAT ',
        }



"""
Tests the fancy indexing and slicing capabilities of Structure
"""
from collections import defaultdict
import numpy as np
import parmed as pmd
import parmed.unit as u
import random
import unittest
import utils
from utils import get_fn

parm = pmd.load_file(get_fn('trx.prmtop'))
pdb1 = pmd.load_file(get_fn('4lzt.pdb'))
pdb2 = pmd.load_file(get_fn('1kip.cif'))
parmep = pmd.load_file(get_fn('tip4p.parm7'))

class TestStructureSlicing(unittest.TestCase):
    """ Tests the fancy slicing/indexing of Structure """

    def test_int_index(self):
        """ Tests simple Structure indexing (integer) """
        for i, atom in enumerate(parm.atoms):
            self.assertIs(atom, parm[i])
        for i, atom in enumerate(pdb1.atoms):
            self.assertIs(atom, pdb1[i])
        for i, atom in enumerate(pdb2.atoms):
            self.assertIs(atom, pdb2[i])

    def test_ub_select(self):
        """ Tests selection when "blank" Urey-Bradley terms present """
        struct = utils.create_random_structure(parametrized=True)
        struct.urey_bradleys[0].type = pmd.NoUreyBradley
        s = struct[:]
        self.assertEqual(len(s.urey_bradleys), len(struct.urey_bradleys))
        self.assertIs(s.urey_bradleys[0].type, pmd.NoUreyBradley)

    def test_two_int_index(self):
        """ Tests simple Structure indexing w/ residue and atom (int, int) """
        for i, res in enumerate(parm.residues):
            for j, atom in enumerate(res):
                self.assertIs(res[j], parm[i,j])
        for i, res in enumerate(pdb1.residues):
            for j, atom in enumerate(res):
                self.assertIs(res[j], pdb1[i,j])
        for i, res in enumerate(pdb2.residues):
            for j, atom in enumerate(res):
                self.assertIs(res[j], pdb2[i,j])

    def test_three_simple_index(self):
        """ Tests simple indexing with chain, residue, and atom ID """
        chains = defaultdict(pmd.TrackedList)
        for res in pdb2.residues:
            chains[res.chain].append(res)
        for chain_name, chain in chains.items():
            for i, res in enumerate(chain):
                for j, atom in enumerate(res):
                    self.assertIs(atom, pdb2[chain_name, i, j])
        self.assertRaises(IndexError, lambda: pdb2['Z', 0, 0])

    def test_single_atom_slice(self):
        """ Test that non-trivial, single-atom selections give a Structure """
        sel1 = parm['@1']
        sel2 = parm[:1]
        self.assertIsInstance(sel1, pmd.Structure)
        self.assertIsInstance(sel2, pmd.Structure)
        self.assertEqual(len(sel1.atoms), 1)
        self.assertEqual(len(sel2.atoms), 1)

    def test_empty_structure(self):
        """ Tests empty selection of a Structure (view) """
        sel1 = parm[10000000:]
        sel2 = pdb1['@NOTHING']
        sel3 = pdb2['G',:,:]
        sel4 = parm['NOTHING',:]
        self.assertFalse(sel1)
        self.assertFalse(sel2)
        self.assertFalse(sel3)
        self.assertFalse(sel4)
        self.assertIsInstance(sel1, pmd.amber.AmberParm)
        self.assertIsInstance(sel2, pmd.Structure)
        self.assertIsInstance(sel3, pmd.Structure)
        self.assertIsInstance(sel4, pmd.amber.AmberParm)

    def test_simple_slice(self):
        """ Tests simple atom slicing """
        sl11 = parm[:10]
        sl21 = pdb1[:10]
        sl31 = pdb2[:10]
        sl12 = parm[10:20]
        sl22 = pdb1[10:20]
        sl32 = pdb2[10:20]
        sl13 = parm[5:100:5] # 19 selections -- 100 not inclusive
        sl23 = pdb1[5:100:5]
        sl33 = pdb2[5:100:5]
        # Check that the slices are all the correct # of atoms
        self.assertEqual(len(sl11.atoms), 10)
        self.assertEqual(len(sl21.atoms), 10)
        self.assertEqual(len(sl31.atoms), 10)
        self.assertEqual(len(sl12.atoms), 10)
        self.assertEqual(len(sl22.atoms), 10)
        self.assertEqual(len(sl32.atoms), 10)
        self.assertEqual(len(sl13.atoms), 19)
        self.assertEqual(len(sl23.atoms), 19)
        self.assertEqual(len(sl33.atoms), 19)

        # Check that the resulting types of the slices are correct
        self.assertIsInstance(sl11, pmd.amber.AmberParm)
        self.assertIsInstance(sl21, pmd.Structure)
        self.assertIsInstance(sl31, pmd.Structure)
        self.assertIsInstance(sl12, pmd.amber.AmberParm)
        self.assertIsInstance(sl22, pmd.Structure)
        self.assertIsInstance(sl32, pmd.Structure)
        self.assertIsInstance(sl13, pmd.amber.AmberParm)
        self.assertIsInstance(sl23, pmd.Structure)
        self.assertIsInstance(sl33, pmd.Structure)

        # Check that the atoms sliced out are correct
        for atom in sl11.atoms:
            patom = parm.atoms[atom.idx]
            self.assertEqual(patom.name, atom.name)
            self.assertEqual(patom.type, atom.type)
            self.assertEqual(patom.charge, atom.charge)
            self.assertEqual(patom.nb_idx, atom.nb_idx)
            self.assertEqual(patom.rmin, atom.rmin)
            self.assertEqual(patom.epsilon, atom.epsilon)
            self.assertEqual(patom.sigma, atom.sigma)

        for atom in sl22.atoms:
            patom = pdb1.atoms[atom.idx+10]
            self.assertEqual(patom.name, atom.name)
            self.assertEqual(patom.xx, atom.xx)
            self.assertEqual(patom.xy, atom.xy)
            self.assertEqual(patom.xz, atom.xz)

        for atom in sl33.atoms:
            patom = pdb2.atoms[atom.idx*5+5]
            self.assertEqual(patom.name, atom.name)
            self.assertEqual(patom.xx, atom.xx)
            self.assertEqual(patom.xy, atom.xy)
            self.assertEqual(patom.xz, atom.xz)

    def test_mask_array(self):
        """ Tests Structure selection using a mask array """
        mask = [a.name in ('CA', 'CB') for a in parm.atoms]
        sel = parm[mask]
        self.assertEqual(len(sel.atoms), 207)
        for atom in sel.atoms:
            self.assertIn(atom.name, ('CA', 'CB'))

    def test_selection_array(self):
        """ Tests Structure selection using indices array """
        indices = [random.randint(0, len(pdb1.atoms)-1) for i in range(100)]
        sel = pdb1[indices]
        self.assertEqual(len(sel.atoms), len(set(indices)))
        self.assertGreater(len(sel.atoms), 0)
        for i, val in enumerate(sorted(set(indices))):
            self.assertEqual(sel.atoms[i].name, pdb1.atoms[val].name)
            self.assertEqual(sel.atoms[i].xx, pdb1.atoms[val].xx)
            self.assertEqual(sel.atoms[i].xy, pdb1.atoms[val].xy)
            self.assertEqual(sel.atoms[i].xz, pdb1.atoms[val].xz)

    def test_residue_atom_selection(self):
        """ Tests combined residue,atom slicing/selections """
        sel = pdb1[10:20,:5] # First five atoms of residues 10-19
        self.assertEqual(len(sel.atoms), 49) # 1 residue has only 4 atoms
        c = 0
        for res in pdb1.residues[10:20]:
            for i in range(min(len(res), 5)):
                self.assertEqual(sel.atoms[c].name, res[i].name)
                self.assertEqual(sel.atoms[c].xx, res[i].xx)
                self.assertEqual(sel.atoms[c].xy, res[i].xy)
                self.assertEqual(sel.atoms[c].xz, res[i].xz)
                c += 1

        sel = pdb1[[0,2,3,5], :2] # First 2 atoms of residues 0, 2, 3, and 5
        self.assertEqual(len(sel.atoms), 8)
        c = 0
        for i, res in enumerate([0, 2, 3, 5]):
            res = pdb1.residues[res]
            self.assertEqual(sel.atoms[c].name, res[0].name)
            self.assertEqual(sel.atoms[c].xx, res[0].xx)
            self.assertEqual(sel.atoms[c].xy, res[0].xy)
            self.assertEqual(sel.atoms[c].xz, res[0].xz)
            c += 1
            self.assertEqual(sel.atoms[c].name, res[1].name)
            self.assertEqual(sel.atoms[c].xx, res[1].xx)
            self.assertEqual(sel.atoms[c].xy, res[1].xy)
            self.assertEqual(sel.atoms[c].xz, res[1].xz)
            c += 1

        sel = pdb1[8,:]
        self.assertEqual(len(sel.atoms), len(pdb1.residues[8]))
        for a1, a2 in zip(sel.atoms, pdb1.residues[8]):
            self.assertEqual(a1.name, a2.name)
            self.assertEqual(a1.xx, a2.xx)
            self.assertEqual(a1.xy, a2.xy)
            self.assertEqual(a1.xz, a2.xz)

        self.assertIs(pdb1[8,4], pdb1.residues[8][4])

        sel = pdb1[['ALA', 'GLY'], :]
        for atom in sel.atoms:
            self.assertIn(atom.residue.name, ('ALA', 'GLY'))

        nalagly = sum(r.name in ('ALA', 'GLY') for r in pdb1.residues)
        natom = sum(len(r) for r in pdb1.residues if r.name in ('ALA', 'GLY'))
        self.assertEqual(nalagly, len(sel.residues))
        self.assertEqual(natom, len(sel.atoms))

    def test_chain_residue_atom_selection(self):
        """ Tests combined chain,residue,atom slicing/selections """
        sel = pdb2['A',:,:] # All of chain A
        for a in sel.atoms:
            self.assertEqual(a.residue.chain, 'A')
        # Get all chains
        chainA = pdb2['A',:,:]
        chainB = pdb2['B',:,:]
        chainC = pdb2['C',:,:]
        for r in chainB.residues:
            self.assertEqual(r.chain, 'B')
        for r in chainC.residues:
            self.assertEqual(r.chain, 'C')
        self.assertEqual(len(pdb2.atoms),
                         len(chainA.atoms)+len(chainB.atoms)+len(chainC.atoms))
        chainAB = pdb2[:2,:,:]
        self.assertEqual(len(chainAB.atoms),
                         len(chainA.atoms)+len(chainB.atoms))
        self.assertFalse(pdb2[10:,:,:])
        # Now try some different kinds of indexing
        sel = pdb2['B',0,0]
        self.assertIs(sel, pdb2.atoms[826])
        sel = pdb2[['A','B'], :10:2, 0]
        self.assertEqual(len(sel.atoms), 2*5*1)
        self.assertEqual(len(pdb2[:,0,[0,1]].atoms), 6) # 3 total chains
        # Now try a bad selector
        self.assertRaises(ValueError, lambda:
                pdb2[list(range(len(pdb2.atoms)+1))])
        self.assertRaises(ValueError, lambda: pdb2[[0,len(pdb2.atoms)]])

    def test_structure_box_and_space_group_and_symmetry(self):
        """ Test correctly copying box, space group and symmetry """
        def assert_correctly_copy(parm):
            sliced_parm = parm['@1-3']
            if parm.box is None:
                self.assertIs(sliced_parm.box, None)
            else:
                np.testing.assert_equal(parm.box, sliced_parm.box)
            if parm.symmetry is None:
                self.assertIs(sliced_parm.symmetry, None)
            else:
                np.testing.assert_equal(parm.symmetry.data, sliced_parm.symmetry.data)
            self.assertEqual(parm.space_group, sliced_parm.space_group)

        # pdb
        parm = pmd.load_file(get_fn('4lzt.pdb'))
        assert_correctly_copy(parm)
        # pdb from rcsb
        parm = pmd.load_file(get_fn('2igd.pdb'))
        assert_correctly_copy(parm)
        # cif
        parm = pmd.load_file(get_fn('sample.cif'))
        assert_correctly_copy(parm)
        # LES parm7, no rst7
        parm = pmd.load_file(get_fn('4lzt.les.parm7'))
        assert_correctly_copy(parm)
        # LES parm7, with rst7
        parm = pmd.load_file(get_fn('4lzt.les.parm7'), xyz=get_fn('4lzt.les.rst7'))
        assert_correctly_copy(parm)
        # parm7, no box
        parm = pmd.load_file(get_fn('ash.parm7'))
        assert_correctly_copy(parm)
        # gro
        parm = pmd.load_file(get_fn('1aki.charmm27.solv.gro'))
        assert_correctly_copy(parm)
        # psf
        parm = pmd.load_file(get_fn('4TVP-dmj_wat-ion.psf'))
        assert_correctly_copy(parm)
        # mol2
        parm = pmd.load_file(get_fn('m2-c1_f3.mol2'), structure=True)
        assert_correctly_copy(parm)
        # pqr
        parm = pmd.load_file(get_fn('adk_open.pqr'))
        assert_correctly_copy(parm)
        # chamber parm
        parm = pmd.load_file(get_fn('ala_ala_ala.parm7'))
        assert_correctly_copy(parm)

class TestStructureViewSlicing(unittest.TestCase):
    """ Tests the fancy slicing/indexing of Structure """

    def test_int_index(self):
        """ Tests simple Structure indexing (integer) (view) """
        for i, atom in enumerate(parm.atoms):
            self.assertIs(atom, parm.view[i])
        for i, atom in enumerate(pdb1.atoms):
            self.assertIs(atom, pdb1.view[i])
        for i, atom in enumerate(pdb2.atoms):
            self.assertIs(atom, pdb2.view[i])

    def test_two_int_index(self):
        """ Tests simple Structure view w/ residue and atom (int, int) """
        for i, res in enumerate(parm.residues):
            for j, atom in enumerate(res):
                self.assertIs(res[j], parm.view[i,j])
        for i, res in enumerate(pdb1.residues):
            for j, atom in enumerate(res):
                self.assertIs(res[j], pdb1.view[i,j])
        for i, res in enumerate(pdb2.residues):
            for j, atom in enumerate(res):
                self.assertIs(res[j], pdb2.view[i,j])

    def test_three_simple_index(self):
        """ Tests simple indexing with chain, residue, and atom ID (view) """
        chains = defaultdict(pmd.TrackedList)
        for res in pdb2.residues:
            chains[res.chain].append(res)
        for chain_name, chain in chains.items():
            for i, res in enumerate(chain):
                for j, atom in enumerate(res):
                    self.assertIs(atom, pdb2.view[chain_name, i, j])

    def test_single_atom_slice(self):
        """ Test that non-trivial, single-atom selections give StructureView """
        sel1 = parm.view['@1']
        sel2 = parm.view[:1]
        self.assertIsInstance(sel1, pmd.structure.StructureView)
        self.assertIsInstance(sel2, pmd.structure.StructureView)
        self.assertEqual(len(sel1.atoms), 1)
        self.assertEqual(len(sel2.atoms), 1)
        self.assertIs(sel1.atoms[0], parm.atoms[0])
        self.assertIs(sel2.atoms[0], parm.atoms[0])

    def test_empty_structure(self):
        """ Tests empty selection of a Structure (view) """
        sel1 = parm.view[10000000:]
        sel2 = pdb1.view['@NOTHING']
        sel3 = pdb2.view['G',:,:]
        sel4 = parm.view['NOTHING',:]
        self.assertFalse(sel1)
        self.assertFalse(sel2)
        self.assertFalse(sel3)
        self.assertFalse(sel4)
        self.assertIsInstance(sel1, pmd.structure.StructureView)
        self.assertIsInstance(sel2, pmd.structure.StructureView)
        self.assertIsInstance(sel3, pmd.structure.StructureView)
        self.assertIsInstance(sel4, pmd.structure.StructureView)

    def test_simple_slice(self):
        """ Tests simple atom slicing (view) """
        sl11 = parm.view[:10]
        sl21 = pdb1.view[:10]
        sl31 = pdb2.view[:10]
        sl12 = parm.view[10:20]
        sl22 = pdb1.view[10:20]
        sl32 = pdb2.view[10:20]
        sl13 = parm.view[5:100:5] # 19 selections -- 100 not inclusive
        sl23 = pdb1.view[5:100:5]
        sl33 = pdb2.view[5:100:5]
        # Check that the slices are all the correct # of atoms
        self.assertEqual(len(sl11.atoms), 10)
        self.assertEqual(len(sl21.atoms), 10)
        self.assertEqual(len(sl31.atoms), 10)
        self.assertEqual(len(sl12.atoms), 10)
        self.assertEqual(len(sl22.atoms), 10)
        self.assertEqual(len(sl32.atoms), 10)
        self.assertEqual(len(sl13.atoms), 19)
        self.assertEqual(len(sl23.atoms), 19)
        self.assertEqual(len(sl33.atoms), 19)

        # Check that the resulting types of the slices are correct
        self.assertIsInstance(sl11, pmd.structure.StructureView)
        self.assertIsInstance(sl21, pmd.structure.StructureView)
        self.assertIsInstance(sl31, pmd.structure.StructureView)
        self.assertIsInstance(sl12, pmd.structure.StructureView)
        self.assertIsInstance(sl22, pmd.structure.StructureView)
        self.assertIsInstance(sl32, pmd.structure.StructureView)
        self.assertIsInstance(sl13, pmd.structure.StructureView)
        self.assertIsInstance(sl23, pmd.structure.StructureView)
        self.assertIsInstance(sl33, pmd.structure.StructureView)

        # Check that the atoms sliced out are correct
        for atom in sl11.atoms:
            self.assertIs(atom, parm.atoms[atom.idx])
        for atom in sl12.atoms:
            self.assertIs(atom, parm.atoms[atom.idx])
        for atom in sl13.atoms:
            self.assertIs(atom, parm.atoms[atom.idx])
        for atom in sl21.atoms:
            self.assertIs(atom, pdb1.atoms[atom.idx])
        for atom in sl22.atoms:
            self.assertIs(atom, pdb1.atoms[atom.idx])
        for atom in sl23.atoms:
            self.assertIs(atom, pdb1.atoms[atom.idx])
        for atom in sl31.atoms:
            self.assertIs(atom, pdb2.atoms[atom.idx])
        for atom in sl32.atoms:
            self.assertIs(atom, pdb2.atoms[atom.idx])
        for atom in sl33.atoms:
            self.assertIs(atom, pdb2.atoms[atom.idx])

        # Check that iterations over views go over the atoms
        for atom in sl11:
            self.assertIs(atom, parm.atoms[atom.idx])

    def test_mask_array(self):
        """ Tests Structure selection using a mask array (view) """
        mask = [a.name in ('CA', 'CB') for a in parm.atoms]
        sel = parm.view[mask]
        self.assertIsInstance(sel, pmd.structure.StructureView)
        self.assertEqual(len(sel.atoms), 207)
        for atom in sel.atoms:
            self.assertIn(atom.name, ('CA', 'CB'))

    def test_selection_array(self):
        """ Tests Structure selection using indices array (view) """
        indices = [random.randint(0, len(pdb1.atoms)-1) for i in range(100)]
        sel = pdb1.view[indices]
        self.assertEqual(len(sel.atoms), len(set(indices)))
        self.assertGreater(len(sel.atoms), 0)
        for i, val in enumerate(sorted(set(indices))):
            self.assertIs(sel.atoms[i], pdb1.atoms[sel.atoms[i].idx])

    def test_residue_atom_selection(self):
        """ Tests combined residue,atom slicing/selections (view) """
        sel = pdb1.view[10:20,:5] # First five atoms of residues 10-19
        self.assertIsInstance(sel, pmd.structure.StructureView)
        self.assertEqual(len(sel.atoms), 49) # 1 residue has only 4 atoms
        c = 0
        for res in pdb1.residues[10:20]:
            for i in range(min(len(res), 5)):
                self.assertIs(sel.atoms[c], res[i])
                c += 1

        sel = pdb1.view[[0,2,3,5], :2] # First 2 atoms of residues 0, 2, 3, and 5
        self.assertIsInstance(sel, pmd.structure.StructureView)
        self.assertEqual(len(sel.atoms), 8)
        c = 0
        for i, res in enumerate([0, 2, 3, 5]):
            res = pdb1.residues[res]
            self.assertIs(sel.atoms[c], res[0])
            c += 1
            self.assertIs(sel.atoms[c], res[1])
            c += 1

        sel = pdb1.view[8,:]
        self.assertIsInstance(sel, pmd.structure.StructureView)
        self.assertEqual(len(sel.atoms), len(pdb1.residues[8]))
        for a1, a2 in zip(sel.atoms, pdb1.residues[8]):
            self.assertIs(a1, a2)

        self.assertIs(pdb1[8,4], pdb1.residues[8][4])

        sel = pdb1.view[['ALA', 'GLY'], :]
        self.assertIsInstance(sel, pmd.structure.StructureView)
        for atom in sel.atoms:
            self.assertIn(atom.residue.name, ('ALA', 'GLY'))
            self.assertIs(atom.list, pdb1.atoms)

        nalagly = sum(r.name in ('ALA', 'GLY') for r in pdb1.residues)
        natom = sum(len(r) for r in pdb1.residues if r.name in ('ALA', 'GLY'))
        self.assertEqual(nalagly, len(sel.residues))
        self.assertEqual(natom, len(sel.atoms))

    def test_chain_residue_atom_selection(self):
        """ Tests combined chain,residue,atom slicing/selections (view) """
        sel = pdb2.view['A',:,:] # All of chain A
        self.assertIsInstance(sel, pmd.structure.StructureView)
        for a in sel.atoms:
            self.assertEqual(a.residue.chain, 'A')
            self.assertIs(a.list, pdb2.atoms)
        # Get all chains
        chainA = pdb2.view['A',:,:]
        chainB = pdb2.view['B',:,:]
        chainC = pdb2.view['C',:,:]
        # Now try some different kinds of indexing
        sel = pdb2.view['B',0,0]
        self.assertIs(sel, pdb2.atoms[826])
        sel = pdb2.view[['A','B'], :10:2, 0]
        self.assertIsInstance(sel, pmd.structure.StructureView)
        self.assertEqual(len(sel.atoms), 2*5*1)

    def test_structure_view_coordinates(self):
        """ Tests handling of coordinates on a StructureView """
        s = utils.create_random_structure(parametrized=True)
        self.assertIs(s.view[:len(s.atoms)//2].coordinates, None)
        self.assertIs(s.view[:len(s.atoms)//2].positions, None)
        # Make sure it's an even number of atoms
        if len(s.atoms) % 2 == 1:
            s.strip('@1')
        x = np.random.rand(len(s.atoms)//2, 3)
        s.coordinates = np.vstack([x.flatten(), x.flatten()])
        np.testing.assert_equal(s.view[:len(s.atoms)//2].coordinates, x)
        np.testing.assert_equal(
                s.view[:len(s.atoms)//2].positions.value_in_unit(u.angstroms), x)

    def test_structure_view_repr(self):
        """ Make sure the __repr__ method works on StructureView """
        repr(parm.view[:10])
        repr(parmep.view[:10])



"""
This module contains all of the ewald namelist variables for the     
different amber programs and automatically loads those dictionaries  
with the default values found in that program (sander or pmemd).     
"""
from typing import Dict

from .typing import InputDataType

class ewald:

    def __init__(self):
        self.sander: Dict[str, InputDataType] = {
            'dsum_tol' : 1.0e-5, 'ew_coeff' : 0.0,
            'skinnb' : 0.0, 'diptol' : 1.0e-4, 'dipmass' : 0.33, 'diptau' : 11.0,
            'nfft1' : 0, 'nfft2' : 0, 'nfft3' : 0, 'order' : 4, 'opt_infl' : 1,
            'verbose' : 0, 'nbflag' : 0, 'nbtell' : 0, 'netfrc' : 12344321,
            'ew_type' : 0, 'vdwmeth' : 1, 'eedmeth' : 1, 'ee_type' : 0,
            'eedtbdns' : 5000.0, 'rsum_tol' : 5.0e-5, 'maxexp' : 0.0,
            'mlimit' : "0,0,0", 'use_pme' : 0, 'maxiter' : 20, 'indmeth' : 3, 'irstdip' : 0,
            'nquench' : 0, 'frameon' : 1, 'chngmask' : 1, 'scaldip' : 1,
            'gridpointers' : 1, 'column_fft' : 1,
        }

        self.pmemd: Dict[str, InputDataType] = {
            'nfft1' : 0, 'nfft2' : 0, 'nfft3' : 0, 'order' : 4, 
            'verbose' : 0, 'ew_type' : 0, 'dsum_tol' : 1.0e-5, 'rsum_tol' : 5.0e-5,
            'mlimit' : "0,0,0", 'ew_coeff' : 0.0, 'nbflag' : 1, 'skinnb' : 2,
            'nbtell' : 0, 'netfrc' : -1, 'use_pme' : 1, 'vdwmeth' : 1,
            'eedmeth' : 1, 'eedtbdns' : 5000.0, 'ee_type' : 0, 'frameon' : 1,
            'chngmask' : 1, 'alpha' : 0.0, 'beta' : 0.0, 'gamma' : 0.0,
            'a' : 0.0, 'b' : 0.0, 'c' : 0.0, 'use_axis_opt' : -1,
            'fft_grids_per_ang' : 1.0, 'block_fft' : -1, 'fft_blk_y_divisor' : -1,
            'excl_recip' : -1, 'excl_master' : -1, 'atm_redist_freq' : -1,
        }



"""
Tests the generalized open routine that is used for virtually all I/O in ParmEd,
and automatically handles compression, etc.
"""

import bz2
import gzip
from contextlib import closing
from parmed.utils.io import genopen
import unittest
from utils import get_fn, FileIOTestCase, is_local

ALPHABET = 'abcdefghijklmnopqrstuvwxyz\n'

class TestGenopen(FileIOTestCase):

    def test_read_normal(self):
        """ Tests genopen reading a normal text file """
        with closing(genopen(get_fn('4lzt.pdb'))) as f:
            self.assertEqual(f.read(), open(get_fn('4lzt.pdb')).read())
        with closing(genopen(get_fn('4lzt.pdb'), 'r')) as f:
            self.assertEqual(f.read(), open(get_fn('4lzt.pdb')).read())

    def test_write_normal(self):
        """ Tests genopen writing a normal text file """
        with closing(genopen(self.get_fn('tmp.txt', written=True), 'w')) as f:
            f.write(ALPHABET)
        self.assertEqual(open(self.get_fn('tmp.txt', written=True), 'r').read(), ALPHABET)

    def test_read_gzipped(self):
        """ Tests genopen reading a gzipped file """
        with closing(genopen(get_fn('4lzt.pdb.gz'))) as f:
            text = gzip.open(get_fn('4lzt.pdb.gz'), 'r').read()
            self.assertEqual(f.read(), text.decode('ascii'))

    def test_write_gzipped(self):
        """ Tests genopen writing a gzipped file """
        with closing(genopen(self.get_fn('test.gz', written=True), 'w')) as f:
            f.write(ALPHABET)
        text = gzip.open(self.get_fn('test.gz', written=True), 'r').read()
        self.assertEqual(text.decode('ascii'), ALPHABET)

    def test_read_bzipped(self):
        """ Tests genopen reading a bzipped file """
        with closing(genopen(get_fn('4lzt.pdb.bz2'), 'r')) as f:
            text = bz2.BZ2File(get_fn('4lzt.pdb.bz2'), 'r').read()
            self.assertEqual(text.decode('ascii'), f.read())

    def test_write_bzipped(self):
        """ Tests genopen writing a bzipped file """
        with closing(genopen(self.get_fn('test.bz2', written=True), 'w')) as f:
            f.write(ALPHABET)
        text = bz2.BZ2File(self.get_fn('test.bz2', written=True), 'r').read()
        self.assertEqual(text.decode('ascii'), ALPHABET)

    def test_read_normal_URL(self):
        """ Tests genopen reading a remote file """
        url = 'https://github.com/ParmEd/ParmEd/raw/master/test/files/tripos1.mol2'
        with closing(genopen(url, 'r')) as f:
            self.assertEqual(f.read(), open(get_fn('tripos1.mol2')).read())

    def test_read_bzipped_URL(self):
        """ Tests genopen reading a bzipped remote file """
        url = 'https://github.com/ParmEd/ParmEd/raw/master/test/files/4lzt.pdb.bz2'
        with closing(genopen(url, 'r')) as f:
            self.assertEqual(f.read(), genopen(get_fn('4lzt.pdb.bz2')).read())

    def test_read_gzipped_URL(self):
        """ Tests genopen reading a gzipped remote file """
        url = 'https://github.com/ParmEd/ParmEd/raw/master/test/files/4lzt.pdb.gz'
        with closing(genopen(url, 'r')) as f:
            self.assertEqual(f.read(), genopen(get_fn('4lzt.pdb.gz')).read())

    # @unittest.skipUnless(is_local(), 'Cannot download files from PDB with CI systems')
    @unittest.skip('RCSB deprecated FTP protocol https://www.rcsb.org/news/feature/65562f0ad78e004e766a96c1')
    def test_read_ftp_URL(self):
        """ Tests genopen reading a ftp remote file """
        url = 'ftp://ftp.wwpdb.org/pub/pdb/data/structures/divided/mmCIF/05/205l.cif.gz'
        with closing(genopen(url, 'r')) as f:
            self.assertEqual(f.read(), genopen(get_fn('205l.cif.gz')).read())

    def test_append_normal(self):
        """ Tests genopen appending a normal text file """
        with closing(genopen(self.get_fn('test.txt', written=True), 'w')) as f:
            f.write(ALPHABET)
        with closing(genopen(self.get_fn('test.txt', written=True), 'a')) as f:
            f.write(ALPHABET)
        self.assertEqual(open(self.get_fn('test.txt', written=True)).read(), ALPHABET*2)

    def test_append_gzip(self):
        """ Tests genopen appending a gzipped file """
        with closing(genopen(self.get_fn('test.txt.gz', written=True), 'a')) as f:
            f.write(ALPHABET)
        with closing(genopen(self.get_fn('test.txt.gz', written=True), 'a')) as f:
            f.write(ALPHABET)
        text = gzip.open(self.get_fn('test.txt.gz', written=True)).read()
        self.assertEqual(text.decode('ascii'), ALPHABET*2)

    def test_append_bzip(self):
        """ Tests genopen appending a bzipped file """
        with closing(genopen(self.get_fn('test.txt.bz2', written=True), 'a')) as f:
            f.write(ALPHABET)
        with closing(genopen(self.get_fn('test.txt.bz2', written=True), 'a')) as f:
            f.write(ALPHABET)
        text = bz2.BZ2File(self.get_fn('test.txt.bz2', written=True)).read()
        self.assertEqual(text.decode('ascii'), ALPHABET*2)

    def test_append_remote_file(self):
        """ Tests that genopen appending a remote file fails """
        url = 'http://q4md-forcefieldtools.org/REDDB/projects/W-73/tripos1.mol2'
        self.assertRaises(ValueError, lambda: genopen(url, 'a'))
        self.assertRaises(ValueError, lambda: genopen(url, 'w'))

    def test_read_bad_URL(self):
        """ Tests proper exception handling of non-existent URL """
        self.assertRaises(IOError, lambda: genopen('http://asdkfjasdf.lib'))




"""
Contains classes for parsing DLPOLY topology and parameter files
"""
__all__ = ['DlpolyFieldFile', 'DlpolyConfigFile']

from ..dlpoly.dlpolyfield import DlpolyFieldFile
from ..dlpoly.dlpolyconfig import DlpolyConfigFile



"""
Tests the functionality in the parmed.gromacs package
"""
from utils import get_fn
from io import StringIO
from parmed.exceptions import PreProcessorError, PreProcessorWarning
from parmed.gromacs._cpp import CPreProcessor
import os
import unittest
import warnings

class TestGromacsCpp(unittest.TestCase):
    """ Tests the C-preprocessor written for the Gromacs classes """

    def test_ifdef(self):
        """ Tests CPreProcessor #ifdef/#else/#endif preprocessing """
        f = StringIO('#ifdef MY_DEFINE\nPPVAR is set to MY_DEFINE\n'
                     '#else\nMY_DEFINE is not set\n#endif\n')
        pp = CPreProcessor(f) # no defines
        self.assertEqual(pp.read().strip(), 'MY_DEFINE is not set')
        f.seek(0)
        self.assertEqual(f.tell(), 0)
        self.assertEqual(pp.tell(), 0)
        pp = CPreProcessor(f, defines=dict(MY_DEFINE='SUCCESS'))
        self.assertEqual(pp.read().strip(), 'PPVAR is set to SUCCESS')

    def test_not_implemented(self):
        """ Test error catching for PP features not implemented """
        f = StringIO('#define MYVAR 1\n#if (MYVAR == 1)\nCORRECT\n#endif')
        pp = CPreProcessor(f)
        self.assertRaises(NotImplementedError, lambda: pp.seek(10))
        self.assertRaises(NotImplementedError, pp.read)
        f = StringIO('#ifdef MYVAR\n#elif defined(NOTMYVAR)\n#else\nline\n#endif')
        pp = CPreProcessor(f)
        self.assertRaises(NotImplementedError, pp.read)

    def test_ifndef(self):
        """ Tests CPreProcessor #ifndef/#else#endif preprocessing """
        f = StringIO('#ifndef MY_DEFINE\nMY_DEFINE is not set\n'
                     '#else\nPPVAR is set to MY_DEFINE\n#endif\n')
        pp = CPreProcessor(f) # no defines
        self.assertEqual(pp.read().strip(), 'MY_DEFINE is not set')
        f.seek(0)
        pp = CPreProcessor(f, defines=dict(MY_DEFINE='SUCCESS'))
        self.assertEqual(pp.read().strip(), 'PPVAR is set to SUCCESS')


    def test_define(self):
        """ Tests CPreProcessor #define preprocessing """
        f = StringIO('#ifdef MY_DEFINE\nPPVAR is set to MY_DEFINE\n'
                     '#else\nMY_DEFINE is not set\n#endif\n'
                     '#define MY_DEFINE SUCCESS\n#ifdef MY_DEFINE\n'
                     'PPVAR is set to MY_DEFINE\n#else\nMY_DEFINE is not set\n'
                     '#endif\n')
        pp = CPreProcessor(f)
        self.assertEqual(pp.read().strip(), 'MY_DEFINE is not set\nPPVAR is set to SUCCESS')
        f = StringIO('#define MYVAR something\nMYVAR\n#define MYVAR something_else\nMYVAR')
        with CPreProcessor(f) as pp, self.assertWarns(PreProcessorWarning):
            pp.read()
        f.seek(0)
        with CPreProcessor(f) as pp:
            self.assertEqual(pp.read().strip(), 'something\nsomething_else')

    def test_undef(self):
        """ Tests CPreProcessor #undef preprocessing """
        # This tests undef-ing both a defined and not-defined variable.
        f = StringIO("MY_DEFINE\n#define MY_DEFINE Changed\nMY_DEFINE\n"
                     "#undef MY_DEFINE\nMY_DEFINE\n#undef NO_VARIABLE\n")
        pp = CPreProcessor(f)
        self.assertEqual(pp.read().strip(), "MY_DEFINE\nChanged\nMY_DEFINE")
        # Make sure undef does not occur where it shouldn't
        f = StringIO('#define MYVAR something\n#ifdef NOVAR\n#undef MYVAR\n'
                     '#endif\nMYVAR')
        with CPreProcessor(f) as pp:
            self.assertEqual(pp.read().strip(), 'something')

    def test_multiword_define(self):
        """ Tests CPreProcessor #define with multiple words """
        f = StringIO("#define MY_DEFINE Changed to a sentence\nMY_DEFINE")
        self.assertEqual(CPreProcessor(f).read().strip(), "Changed to a sentence")

    def test_pp_regex(self):
        """ Tests CPreProcessor processing with whitespace """
        f = StringIO("#      define MY_DEFINE Something\nMY_DEFINE")
        self.assertEqual(CPreProcessor(f).read().strip(), "Something")
        f = StringIO("#       ifdef      MY_VAR    \nMY_VAR\n#endif\n"
                     "#\t\tdefine MY_VAR       NoSpaces    \n"
                     "#       ifdef      MY_VAR    \nMY_VAR\n#endif")
        f.seek(0)
        self.assertEqual(CPreProcessor(f).read().strip(), 'NoSpaces')

    def test_define_inside_ifdef(self):
        """ Tests CPreProcessor #define inside #ifdef/#ifndef """
        f = StringIO("#ifdef MY_DEFINE\n#define MY_DEFINE "
                     "Something\n#endif\nMY_DEFINE")
        pp = CPreProcessor(f)
        self.assertEqual(pp.read().strip(), "MY_DEFINE")
        f = StringIO("#ifdef MY_DEFINE\n#    define MY_DEFINE Something\n"
                     "#endif\nMY_DEFINE\n#ifndef MY_DEFINE\n"
                     "#    define MY_DEFINE Something special\n#endif\n"
                     "MY_DEFINE")
        pp = CPreProcessor(f)
        self.assertEqual(pp.read().strip(), "MY_DEFINE\nSomething special")

    def test_if1_if0(self):
        """ Tests CPreProcessor use of #if 1 and #if 0 to serve as comments """
        f = StringIO('#if 1\n#ifdef MY_DEFINE\nPPVAR is set to MY_DEFINE\n'
                     '#else\nMY_DEFINE is not set\n#endif\n#endif\n')
        pp = CPreProcessor(f) # no defines
        self.assertEqual(pp.read().strip(), 'MY_DEFINE is not set')
        f.seek(0)
        pp = CPreProcessor(f, defines=dict(MY_DEFINE='SUCCESS'))
        self.assertEqual(pp.read().strip(), 'PPVAR is set to SUCCESS')
        f = StringIO('#if 0\n#ifdef MY_DEFINE\nPPVAR is set to MY_DEFINE\n'
                     '#else\nMY_DEFINE is not set\n#endif\n#endif\n')
        pp = CPreProcessor(f) # no defines
        self.assertFalse(pp.read().strip())
        f.seek(0)
        pp = CPreProcessor(f, defines=dict(MY_DEFINE='SUCCESS'))
        self.assertFalse(pp.read().strip())

    def test_nested_ifs(self):
        """ Tests CPreProcessor nested #ifdef and #if """
        f = StringIO("""
#ifdef MY_VAR
line 1
#   if 1
line 2
#   endif /* 1 */
#else /* ! MY_VAR */
#   if 0
line 3
#   else /* ! 0 */
line 4
#   endif /* 0 */
#endif /* MY_VAR */
""")
        pp = CPreProcessor(f)
        self.assertEqual(pp.read().strip(), "line 4")
        f.seek(0)
        pp = CPreProcessor(f, defines=dict(MY_VAR=1))
        self.assertEqual(pp.read().strip(), "line 1\nline 2")
        f = StringIO("""
#ifdef MY_VAR
#   ifndef MY_VAR_2
MY_VAR defined... MY_VAR_2 not
#   else
MY_VAR defined... MY_VAR_2 also
#   endif
#endif
""")
        pp = CPreProcessor(f)
        self.assertEqual(pp.read().strip(), '')
        f.seek(0)
        pp = CPreProcessor(f, defines=dict(MY_VAR=1))
        self.assertEqual(pp.read().strip(), '1 defined... MY_VAR_2 not')
        f.seek(0)
        pp = CPreProcessor(f, defines=dict(MY_VAR=1, MY_VAR_2=2))
        self.assertEqual(pp.read().strip(), '1 defined... 2 also')

    def test_simple_include(self):
        """ Tests CPreProcessor simple #include directive """
        pp = CPreProcessor(get_fn('pptest1/pptest1.h'))
        self.assertEqual(pp.read().strip(), """\
pptest1 line 1
pptest2 line 1
pptest3 line 1
pptest2 line 2
pptest1 line 2
pptest3 line 1
pptest1 line 3""")

    def test_include_defines(self):
        """ Tests CPreProcessor passing defines between included files """
        pp = CPreProcessor(get_fn('pptest2/pptest1.h'))
        self.assertEqual(pp.read().strip(), """\
pptest1 line 1
pptest2 line 1
pptest3 line 1
pptest1 line 2
pptest1 line 3
pptest3 line 1
pptest1 line 4""")

    def test_include_dir(self):
        """ Tests CPreProcessor passing include directories to search """
        f = StringIO('#include "pptest1.h"')
        pp = CPreProcessor(f, includes=[get_fn('pptest1')])
        self.assertEqual(pp.read().strip(), """\
pptest1 line 1
pptest2 line 1
pptest3 line 1
pptest2 line 2
pptest1 line 2
pptest3 line 1
pptest1 line 3""")

    def test_conservative_defines(self):
        """ Tests CPreProcessor #define token replacement """
        # Check that it does not replace tokens inside quotes
        f = StringIO("'MY_VAR' is MY_VAR\n\"MY_VAR\" is MY_VAR still\n")
        pp = CPreProcessor(f, defines=dict(MY_VAR='set'))
        self.assertEqual(pp.read().strip(), "'MY_VAR' is set\n\"MY_VAR\" is set still")
        f = StringIO("This is NOT_MY_VAR, but 'MY_VAR' is MY_VAR\n")
        pp = CPreProcessor(f, defines=dict(MY_VAR='taken'))
        self.assertEqual(pp.read().strip(), "This is NOT_MY_VAR, but 'MY_VAR' is taken")

    def test_nested_defines(self):
        """ Tests CPreProcessor nested #define token replacement """
        f = StringIO("#define MY_VAR replace1\n#define MY_VAR2 MY_VAR\n"
                     "MY_VAR2\n")
        pp = CPreProcessor(f)
        self.assertEqual(pp.read().strip(), 'replace1')

    def test_close(self):
        """ Test closing file object before deleting it """
        pp = CPreProcessor(get_fn('pptest1/pptest1.h'))
        pp.close()
        del pp

    def test_warnings(self):
        """ Test CPreProcessor warnings """
        f = StringIO('#ifndef MYVAR extra tokens\nline\n#endif')
        with CPreProcessor(f) as f, self.assertWarns(PreProcessorWarning):
            f.read()

    def test_context_manager(self):
        """ Test using CPreProcessor in a context manager """
        with CPreProcessor(get_fn('pptest2/pptest1.h')) as pp:
            self.assertEqual(pp.read().strip(), """\
pptest1 line 1
pptest2 line 1
pptest3 line 1
pptest1 line 2
pptest1 line 3
pptest3 line 1
pptest1 line 4""")

    def test_bad_ifdef(self):
        """ Tests CPreProcessor error processing of bad #ifdef/#ifndef """
        f = StringIO("#ifdef\n#endif")
        with CPreProcessor(f) as pp, self.assertRaises(PreProcessorError):
            pp.read()
        f = StringIO("#ifndef\n#endif")
        with CPreProcessor(f) as pp, self.assertRaises(PreProcessorError):
            pp.read()
        f = StringIO("#ifdef SOME_VAR\ndangling ifdef\n\n\n")
        with CPreProcessor(f) as pp, self.assertRaises(PreProcessorError):
            pp.read()
        f = StringIO("#ifdef SOME_VAR misplaced comments\n#endif\n")
        with CPreProcessor(f) as pp, self.assertWarns(PreProcessorWarning):
            pp.read()
        f = StringIO("#ifdef SOME_VAR\n#endif misplaced comments\n")
        with CPreProcessor(f) as pp, self.assertWarns(PreProcessorWarning):
            pp.read()
        f = StringIO('#else\nNOVAR\n#endif')
        with CPreProcessor(f) as pp, self.assertRaises(PreProcessorError):
            pp.read()
        f = StringIO('#ifdef MYVAR\nline1\n#else extra tokens\nline2\n#endif')
        with CPreProcessor(f) as pp, self.assertWarns(PreProcessorWarning):
            pp.read()
        f = StringIO('#ifdef MYVAR\nline1\n#else\nline2\n#else\nline3\n#endif')
        with CPreProcessor(f) as pp, self.assertRaises(PreProcessorError):
            pp.read()
        f = StringIO('#endif\nline1\nline2')
        with CPreProcessor(f) as pp, self.assertRaises(PreProcessorError):
            pp.read()

    def test_bad_define_undef(self):
        """ Tests CPreProcessor error processing of bad #define/#undef """
        f = StringIO("#define\n")
        pp = CPreProcessor(f)
        self.assertRaises(PreProcessorError, lambda: pp.read())
        f = StringIO("#undef\n")
        pp = CPreProcessor(f)
        self.assertRaises(PreProcessorError, lambda: pp.read())
        f = StringIO('#undef VAR1 misplaced comments\n')
        pp = CPreProcessor(f)
        with self.assertWarns(PreProcessorWarning):
            pp.read()

    def test_missing_include(self):
        """ Tests CPreProcessor controllable behavior of missing #includes """
        f = StringIO("#include \"nofile.h\"\n")
        pp = CPreProcessor(f)
        self.assertRaises(PreProcessorError, lambda: pp.read())
        f.seek(0)
        pp = CPreProcessor(f, notfound_fatal=False)
        with self.assertWarns(PreProcessorWarning):
            pp.read()

    def test_bad_include(self):
        """ Tests bad #include syntax in CPreProcessor """
        f = StringIO('#include bad syntax gremlin\nline 1')
        with CPreProcessor(f) as pp, self.assertRaises(PreProcessorError):
            pp.read()



"""
Tests the functionality in parmed.modeller
"""
from copy import copy
from io import StringIO
import numpy as np
try:
    import pandas as pd
except ImportError:
    pd = None
try:
    import networkx as nx
except ImportError:
    nx = None
import os
import pytest
import parmed as pmd
from parmed import Atom, read_PDB, Structure
from parmed.amber import AmberParm, AmberOFFLibrary
from parmed.exceptions import AmberWarning, Mol2Error
from parmed.modeller import (ResidueTemplate, ResidueTemplateContainer, PatchTemplate,
                             PROTEIN, SOLVENT, StandardBiomolecularResidues)
from parmed.formats import Mol2File, PDBFile
from parmed.geometry import distance2
from parmed.exceptions import MoleculeError
from parmed.utils import find_atom_pairs
from parmed.tools import changeRadii
import random
import sys
import unittest
import utils
import warnings
get_fn = utils.get_fn

class TestResidueTemplate(unittest.TestCase):
    """ Tests the ResidueTemplate class """

    def setUp(self):
        self.templ = templ = ResidueTemplate('ACE')
        templ.add_atom(Atom(name='HH31', type='HC', atomic_number=1))
        templ.add_atom(Atom(name='CH3', type='CT', atomic_number=6))
        templ.add_atom(Atom(name='HH32', type='HC', atomic_number=1))
        templ.add_atom(Atom(name='HH33', type='HC', atomic_number=1))
        templ.add_atom(Atom(name='C', type='C', atomic_number=6))
        templ.add_atom(Atom(name='O', type='O', atomic_number=8))
        templ.type = PROTEIN
        templ.tail = templ.atoms[4]
        repr(PROTEIN) # make sure it works, don't care what its value is
        assert str(PROTEIN) == 'PROTEIN'

    def test_chemical_formula(self):
        """ Test computation of empirical chemical formula. """
        a1, a2, a3, a4, a5, a6 = self.templ.atoms
        self.templ.add_bond(a1, a2)
        self.templ.add_bond(a2, a3)
        self.templ.add_bond(a2, a4)
        self.templ.add_bond(a2, a5)
        self.templ.add_bond(a5, a6)
        self.assertEqual(self.templ.empirical_chemical_formula, 'C2H3O')

    @unittest.skipIf(pd is None, "Cannot test without pandas")
    def test_data_frame(self):
        """ Test converting ResidueTemplate to a DataFrame """
        df = self.templ.to_dataframe()
        self.assertEqual(df.shape, (6, 20))
        self.assertAlmostEqual(df.charge.sum(), 0)
        self.assertEqual(df.atomic_number.sum(), 23)

    @unittest.skipIf(nx is None, "Cannot test without networkx")
    def test_to_networkx(self):
        """ Test converting ResidueTemplate to NetworkX graph """
        a1, a2, a3, a4, a5, a6 = self.templ.atoms
        self.templ.add_bond(a1, a2)
        self.templ.add_bond(a2, a3)
        self.templ.add_bond(a3, a4)
        self.templ.add_bond(a5, a6)
        G1 = self.templ.to_networkx()
        assert not nx.is_connected(G1)
        self.templ.add_bond(a2, a5)
        G2 = self.templ.to_networkx()
        assert nx.is_connected(G2)

    def test_add_atom(self):
        """ Tests the ResidueTemplate.add_atom function """
        templ = self.templ
        self.assertEqual(len(templ), 6)
        for i in range(6):
            self.assertIs(templ[i], templ.atoms[i])
        for x, y in zip(templ, templ.atoms):
            self.assertIs(x, y)
        for i, atom in enumerate(templ):
            self.assertIs(atom, templ.atoms[i])
        self.assertIs(templ.head, None)
        self.assertIs(templ.tail, templ[-2])
        self.assertRaises(ValueError, lambda: templ.add_atom(Atom(name='C')))
        self.assertRaises(IndexError, lambda: templ['NOAT'])
        # Make sure we can print an atom's repr when it is in a ResidueTemplate
        repr(templ.atoms[0])

    def test_to_structure(self):
        """ Tests the ResidueTemplate.to_structure function """
        a1, a2, a3, a4, a5, a6 = self.templ.atoms
        self.templ.add_bond(a1, a2)
        self.templ.add_bond(a2, a3)
        self.templ.add_bond(a3, a4)
        self.templ.add_bond(a2, a5)
        self.templ.add_bond(a5, a6)
        struct = self.templ.to_structure()

        self.assertIsInstance(struct, Structure)
        self.assertEqual(len(struct.atoms), 6)
        self.assertEqual(len(struct.residues), 1)
        self.assertEqual(len(struct.bonds), 5)

    def test_delete_bond(self):
        """ Tests the ResidueTemplate.delete_bond function """
        templ = self.templ

        a1, a2, a3, a4, a5, a6 = self.templ.atoms
        self.templ.add_bond(a1, a2)
        self.templ.add_bond(a2, a3)
        self.templ.add_bond(a3, a4)
        self.templ.add_bond(a2, a5)
        self.templ.add_bond(a5, a6)

        bond = templ.bonds[0]
        self.assertEqual(len(templ.bonds), 5)
        templ.delete_bond(bond)
        self.assertEqual(len(templ.bonds), 4)
        # Make sure we can't delete the bond again
        self.assertRaises(ValueError, lambda: templ.delete_bond(bond))

    def test_copy(self):
        """ Tests ResidueTemplate __copy__ functionality """
        for atom in self.templ:
            atom.xx = random.random() * 20 - 10
            atom.xy = random.random() * 20 - 10
            atom.xz = random.random() * 20 - 10
        templcopy = copy(self.templ)
        self.assertIsNot(templcopy, self.templ)
        self.assertEqual(len(templcopy.atoms), len(self.templ.atoms))
        self.assertEqual(len(templcopy.bonds), len(self.templ.bonds))
        for a1, a2 in zip(templcopy.atoms, self.templ):
            self.assertIsNot(a1, a2)
            self.assertEqual(a1.name, a2.name)
            self.assertEqual(a1.charge, a2.charge)
            self.assertEqual(a1.xx, a2.xx)
            self.assertEqual(a1.xy, a2.xy)
            self.assertEqual(a1.xz, a2.xz)
        for b1, b2 in zip(templcopy.bonds, self.templ.bonds):
            self.assertIsNot(b1, b2)
            self.assertIsNot(b1.atom1, b2.atom1)
            self.assertIsNot(b1.atom2, b2.atom2)
            self.assertEqual(b1.atom1.name, b2.atom1.name)
            self.assertEqual(b1.atom2.name, b2.atom2.name)
        self.assertIs(self.templ.head, None)
        self.assertIs(templcopy.head, None)
        self.assertIs(self.templ.tail, self.templ[4])
        self.assertIs(templcopy.tail, templcopy[4])
        self.assertIs(templcopy.type, self.templ.type)
        self.assertEqual(templcopy.name, self.templ.name)

        # Give ResidueTemplate a head atom and a connection
        self.templ.head = self.templ[0]
        self.templ.connections.append(self.templ[1])
        templcopy = copy(self.templ)
        self.assertIs(self.templ.head, self.templ[0])
        self.assertIs(templcopy.head, templcopy[0])
        self.assertEqual(len(self.templ.connections), 1)
        self.assertEqual(len(templcopy.connections), 1)
        for a1, a2 in zip(self.templ.connections, templcopy.connections):
            self.assertIsNot(a1, a2)
            self.assertEqual(a1.name, a2.name) # Enough to verify correctness
        self.assertIs(self.templ.first_patch, None)
        self.assertIs(self.templ.last_patch, None)
        self.assertIs(templcopy.first_patch, None)
        self.assertIs(templcopy.last_patch, None)

        # Now give ResidueTemplate patches (can be blank, doesn't matter)
        self.templ.first_patch = ResidueTemplate()
        self.templ.last_patch = ResidueTemplate()
        templcopy = copy(self.templ)
        self.assertIsNot(self.templ.first_patch, None)
        self.assertIsNot(self.templ.last_patch, None)
        self.assertIs(self.templ.first_patch, templcopy.first_patch)
        self.assertIs(self.templ.last_patch, templcopy.last_patch)

    def test_fix_charge(self):
        """ Tests charge fixing for ResidueTemplate """
        self.assertRaises(ValueError, lambda:
                ResidueTemplate().fix_charges())
        charges = [random.random()*2 - 2 for a in self.templ]
        for a, charge in zip(self.templ, charges):
            a.charge = charge
        net_charge = sum(charges)
        # The odds of 6 random numbers adding to an exact integer is miniscule
        self.assertNotEqual(net_charge, round(net_charge))
        self.assertEqual(round(net_charge), int(round(net_charge)))
        # Find what the new charges *should* be
        diff = (net_charge - round(net_charge)) / 6
        fixed_charges = [x - diff for x in charges]
        self.assertAlmostEqual(sum(fixed_charges), round(net_charge), places=10)
        # Fix the charges
        self.assertEqual(sum(a.charge for a in self.templ), net_charge)
        precision = random.randint(4, 10)
        return_value = self.templ.fix_charges(precision=precision)
        self.assertAlmostEqual(sum(a.charge for a in self.templ),
                               sum(fixed_charges), places=10)
        for a, chg in zip(self.templ, fixed_charges):
            self.assertAlmostEqual(a.charge, chg, delta=2*10**-precision)
        # Check that the return value is the residue itself
        self.assertIs(return_value, self.templ)

    def test_fix_charge_2(self):
        """ Tests charge fixing to a specific value for ResidueTemplate """
        desired_charge = random.choice(range(-10, 11))
        charges = [random.random()*2 - 2 for a in self.templ]
        for a, charge in zip(self.templ, charges):
            a.charge = charge
        net_charge = sum(charges)
        # The odds of 6 random numbers adding to my exact integer are negligible
        self.assertNotEqual(net_charge, desired_charge)
        # Find what the new charges *should* be
        diff = (net_charge - desired_charge) / 6
        fixed_charges = [x - diff for x in charges]
        self.assertAlmostEqual(sum(fixed_charges), desired_charge, places=10)
        # Fix the charges
        self.assertAlmostEqual(sum(a.charge for a in self.templ), net_charge,
                               places=10)
        precision = random.randint(4, 10)
        return_value = self.templ.fix_charges(desired_charge, precision)
        self.assertAlmostEqual(sum(a.charge for a in self.templ),
                               sum(fixed_charges), places=10)
        for a, chg in zip(self.templ, fixed_charges):
            self.assertAlmostEqual(a.charge, chg, delta=2*10**precision)
        # Check that the return value is the residue itself
        self.assertIs(return_value, self.templ)

    def test_fix_charge_container(self):
        """ Tests charge fixing for ResidueTemplateContainer """
        rescont = ResidueTemplateContainer()
        self.assertRaises(ValueError, rescont.fix_charges)
        for i in range(10):
            templcopy = copy(self.templ)
            templcopy.name = '%s%d' % (templcopy.name, i)
            sumchg = 0
            for a in templcopy.atoms:
                a.charge = random.random()*2 - 2
                sumchg += a.charge
            self.assertNotEqual(sumchg, round(sumchg))
            rescont.append(templcopy)
        self.assertEqual(len(rescont), 10)
        orig_charges = [sum(a.charge for a in r) for r in rescont]
        new_charges = [round(x) for x in orig_charges]
        for res, oc, nc in zip(rescont, orig_charges, new_charges):
            self.assertNotEqual(oc, nc)
            self.assertEqual(round(oc), nc)
            self.assertEqual(sum(a.charge for a in res), oc)
        precision = random.randint(4, 10)
        retval = rescont.fix_charges(precision=precision)
        self.assertIs(retval, rescont)
        for res, oc, nc in zip(rescont, orig_charges, new_charges):
            self.assertNotEqual(oc, nc)
            self.assertEqual(round(oc), nc)
            strchgs = ['%%.%df' % precision % atom.charge for atom in res.atoms]
            # If the string charges are equal to the rounded charge to *greater
            # than the requested precision*, then it is clearly exactly equal.
            # We go 3 orders of magnitude tighter than the printed precision to
            # make sure
            self.assertAlmostEqual(round(oc), sum(float(x) for x in strchgs),
                                   places=precision+3)
            self.assertAlmostEqual(sum(a.charge for a in res), nc, places=10)
        # Make all charges alternate between 1.0 and -1.0 for the first residue,
        # then show that fixing charges does not change them
        charges = []
        for i, atom in enumerate(rescont[0].atoms):
            atom.charge = 1**(-i)
            charges.append(1**(-i))
        rescont[0].fix_charges()
        for a, c in zip(rescont[0].atoms, charges):
            self.assertEqual(a.charge, c)

    def test_delete_atom(self):
        """ Tests the ResidueTemplate.delete_atom function """
        templ = copy(self.templ)
        a1, a2, a3, a4, a5, a6 = templ.atoms
        a7 = Atom(name='Unimportant', type='black hole')
        templ.add_bond(a1, a2)
        templ.add_bond(a2, a3)
        templ.add_bond(a3, a4)
        templ.add_bond(a2, a5)
        templ.add_bond(a5, a6)
        templ.delete_atom(a6)
        #self.assertRaises(RuntimeError, lambda: templ.delete_atom(a7))
        self.assertIn(a1, a2.bond_partners)
        self.assertIn(a2, a1.bond_partners)
        self.assertIn(a3, a2.bond_partners)
        self.assertIn(a2, a3.bond_partners)
        self.assertIn(a5, a2.bond_partners)
        self.assertIn(a2, a5.bond_partners)
        self.assertNotIn(a5, a6.bond_partners)
        self.assertNotIn(a6, a5.bond_partners)
        self.assertEqual(len(templ.bonds), 4)

    @unittest.skipIf(nx is None, "Cannot test without networkx")
    def test_patch_residue(self):
        """ Tests ResidueTemplate.patch_residue function """
        templ = self.templ
        a1, a2, a3, a4, a5, a6 = templ.atoms
        templ.add_bond(a1, a2)
        templ.add_bond(a2, a3)
        templ.add_bond(a2, a4)
        templ.add_bond(a2, a5)
        templ.add_bond(a5, a6)
        self.assertIs(templ.tail, templ[4])
        patch = PatchTemplate()
        patch.delete_atoms.append(a5.name)
        patch.delete_atoms.append(a6.name)
        residue = self.templ.apply_patch(patch)
        self.assertEqual(len(residue.atoms), 4)
        self.assertEqual(len(residue.bonds), 3)
        a1, a2, a3, a4 = residue.atoms
        self.assertIn(a1, a2.bond_partners)
        self.assertIn(a2, a1.bond_partners)
        self.assertIn(a3, a2.bond_partners)
        self.assertIn(a2, a3.bond_partners)
        self.assertIn(a4, a2.bond_partners)
        self.assertIn(a2, a4.bond_partners)
        self.assertIs(residue.tail, None)

    def test_add_bonds_atoms(self):
        """ Tests the ResidueTemplate.add_bond function w/ indices """
        templ = self.templ
        a1, a2, a3, a4, a5, a6 = templ.atoms
        a7 = Atom(name='Unimportant', type='black hole')
        templ.add_bond(a1, a2)
        templ.add_bond(a2, a3)
        templ.add_bond(a3, a4)
        templ.add_bond(a2, a5)
        templ.add_bond(a5, a6)
        self.assertRaises(RuntimeError, lambda: templ.add_bond(a1, a7))
        self.assertRaises(MoleculeError, lambda: templ.add_bond(a1, a1))
        self.assertIn(a1, a2.bond_partners)
        self.assertIn(a2, a1.bond_partners)
        self.assertIn(a3, a2.bond_partners)
        self.assertIn(a2, a3.bond_partners)
        self.assertIn(a5, a2.bond_partners)
        self.assertIn(a2, a5.bond_partners)
        self.assertIn(a5, a6.bond_partners)
        self.assertIn(a6, a5.bond_partners)
        self.assertEqual(len(templ.bonds), 5)

    def test_add_bonds_idx(self):
        """ Tests the ResidueTemplate.add_bond function w/ atoms """
        templ = self.templ
        a1, a2, a3, a4, a5, a6 = range(6)
        a7 = Atom(name='Unimportant', type='black hole')
        templ.add_bond(a1, a2)
        templ.add_bond(a2, a3)
        templ.add_bond(a3, a4)
        templ.add_bond(a2, a5)
        templ.add_bond(a5, a6)
        self.assertRaises(RuntimeError, lambda: templ.add_bond(a1, a7))
        self.assertRaises(MoleculeError, lambda: templ.add_bond(a1, a1))
        a1, a2, a3, a4, a5, a6 = templ.atoms
        self.assertIn(a1, a2.bond_partners)
        self.assertIn(a2, a1.bond_partners)
        self.assertIn(a3, a2.bond_partners)
        self.assertIn(a2, a3.bond_partners)
        self.assertIn(a5, a2.bond_partners)
        self.assertIn(a2, a5.bond_partners)
        self.assertIn(a5, a6.bond_partners)
        self.assertIn(a6, a5.bond_partners)
        self.assertEqual(len(templ.bonds), 5)

    def test_from_residue(self):
        """ Tests the ResidueTemplate.from_residue function """
        # Grab this residue from an amber prmtop file
        struct = AmberParm(get_fn('trx.prmtop'), get_fn('trx.inpcrd'))
        for res in struct.residues:
            self._check_arbitrary_res(struct, res)

    def test_from_residue_noorder(self):
        """ Tests the from_residue function when residue order unknown """
        struct = AmberParm(get_fn('trx.prmtop'), get_fn('trx.inpcrd'))
        residues = struct.residues[:]
        del struct.residues[:]
        for res in residues:
            self.assertEqual(res.idx, -1)
        self.assertGreater(len(residues), 0)
        # Check what should be a warning
        with self.assertWarns(UserWarning):
            ResidueTemplate.from_residue(residues[1])

    def _check_arbitrary_res(self, struct, res):
        orig_indices = [a.idx for a in res]
        templ = ResidueTemplate.from_residue(res)
        # Make sure we didn't clobber any of the atoms in res
        for i, atom in zip(orig_indices, res.atoms):
            self.assertIs(atom.list, struct.atoms)
            self.assertEqual(atom.idx, i)
        # Make sure that we have the same number of atoms in the residue as the
        # source
        self.assertEqual(len(res), len(templ))
        for a1, a2 in zip(res, templ):
            self.assertIsInstance(a1, Atom)
            self.assertIsInstance(a2, Atom)
            self.assertEqual(a1.name, a2.name)
            self.assertEqual(a1.type, a2.type)
            self.assertEqual(a1.atomic_number, a2.atomic_number)
            self.assertEqual(a1.xx, a2.xx)
            self.assertEqual(a1.xy, a2.xy)
            self.assertEqual(a1.xz, a2.xz)
            self.assertIn(a2, templ)
            self.assertNotIn(a1, templ)
        # Make sure we have the correct number of bonds in the residue
        bondset = set()
        for atom in res:
            for bond in atom.bonds:
                if bond.atom1 in res and bond.atom2 in res:
                    bondset.add(bond)
        self.assertGreater(len(bondset), 0)
        self.assertEqual(len(bondset), len(templ.bonds))
        # Make sure that each atom has the correct number of bonds
        for i, atom in enumerate(res):
            for bond in atom.bonds:
                try:
                    id1 = res.atoms.index(bond.atom1)
                    id2 = res.atoms.index(bond.atom2)
                except ValueError:
                    if bond.atom1 in res:
                        oatom = bond.atom2
                        idx = res.atoms.index(bond.atom1)
                    else:
                        oatom = bond.atom1
                        idx = res.atoms.index(bond.atom2)
                    if oatom.residue.idx == res.idx - 1:
                        self.assertIs(templ.head, templ[idx])
                    elif oatom.residue.idx == res.idx + 1:
                        self.assertIs(templ.tail, templ[idx])
                    elif oatom.residue.idx == res.idx:
                        self.assertTrue(False) # Should never hit
                    else:
                        # Should only happen with CYX for amber prmtop...
                        self.assertEqual(res.name, 'CYX')
                        self.assertEqual(atom.name, 'SG')
                        if bond.atom1 in res:
                            self.assertIn(templ[idx], templ.connections)
                else:
                    self.assertIn(templ[id1], templ[id2].bond_partners)
                    self.assertIn(templ[id2], templ[id1].bond_partners)
        # Make sure that our coordinates come as a numpy array
        self.assertIsInstance(templ.coordinates, np.ndarray)
        self.assertEqual(templ.coordinates.shape, (len(templ), 3))
        # Check that repr doesn't fail
        repr(templ)

class TestResidueTemplateContainer(unittest.TestCase):
    """ Tests the ResidueTemplateContainer class """

    def test_from_structure(self):
        """ Tests building ResidueTemplateContainer from a Structure """
        struct = AmberParm(get_fn('trx.prmtop'), get_fn('trx.inpcrd'))
        cont = ResidueTemplateContainer.from_structure(struct)
        for res, sres in zip(cont, struct.residues):
            self.assertIsInstance(res, ResidueTemplate)
            self.assertEqual(len(res), len(sres))
            for a1, a2 in zip(res, sres):
                self.assertEqual(a1.name, a2.name)
                self.assertEqual(a1.type, a2.type)
                self.assertEqual(a1.charge, a2.charge)
                self.assertEqual(a1.xx, a2.xx)
                self.assertEqual(a1.xy, a2.xy)
                self.assertEqual(a1.xz, a2.xz)
        # Check accessor
        self.assertIs(cont[0], cont[cont[0].name])

    def test_to_library(self):
        """ Tests converting a ResidueTemplateContainer to a library/dict """
        lib = ResidueTemplateContainer.from_structure(
                AmberParm(get_fn('trx.prmtop'), get_fn('trx.inpcrd'))
        ).to_library()
        self.assertIsInstance(lib, dict)
        self.assertEqual(len(lib.keys()), 23)
        refset = set(["NSER", "ASP", "LYS", "ILE", "HID", "LEU", "THR", "SER",
                      "PHE", "VAL", "ALA", "GLY", "ASH", "TRP", "GLU", "CYX",
                      "PRO", "MET", "TYR", "GLN", "ASN", "ARG", "CALA"])
        self.assertEqual(set(lib.keys()), refset)

class TestResidueTemplateSaver(utils.FileIOTestCase):
    " Tests the .save method on ResidueTemplate and ResidueTemplateContainer "

    def setUp(self):
        # Create a ResidueTemplate
        utils.FileIOTestCase.setUp(self)
        self.ace = templ = ResidueTemplate('ACE')
        templ.add_atom(Atom(name='HH31', type='HC'))
        templ.add_atom(Atom(name='CH3', type='CT'))
        templ.add_atom(Atom(name='HH32', type='HC'))
        templ.add_atom(Atom(name='HH33', type='HC'))
        templ.add_atom(Atom(name='C', type='C'))
        templ.add_atom(Atom(name='O', type='O'))
        templ.add_bond(0, 1)
        templ.add_bond(1, 2)
        templ.add_bond(1, 3)
        templ.add_bond(1, 4)
        templ.add_bond(4, 5)
        templ.type = PROTEIN
        templ.tail = templ.atoms[4]
        for a in templ.atoms:
            a.xx = random.random() * 10 - 5
            a.xy = random.random() * 10 - 5
            a.xz = random.random() * 10 - 5

        self.nme = templ = ResidueTemplate('NME')
        templ.add_atom(Atom(name='N', type='N'))
        templ.add_atom(Atom(name='H', type='H'))
        templ.add_atom(Atom(name='CH3', type='CT'))
        templ.add_atom(Atom(name='H31', type='H1'))
        templ.add_atom(Atom(name='H32', type='H1'))
        templ.add_atom(Atom(name='H33', type='H1'))
        templ.add_bond(0, 1)
        templ.add_bond(0, 2)
        templ.add_bond(2, 3)
        templ.add_bond(2, 4)
        templ.add_bond(2, 5)
        templ.head = templ.atoms[0]
        for a in templ.atoms:
            a.xx = random.random() * 10 - 5
            a.xy = random.random() * 10 - 5
            a.xz = random.random() * 10 - 5

        self.container = ResidueTemplateContainer()
        self.container.append(self.ace)
        self.container.append(self.nme)

    def test_residue_template_mol2_save(self):
        """ Tests ResidueTemplate.save() method for Mol2 file """
        # Check saving mol2 files by keyword
        self.ace.save(self.get_fn('test', written=True), format='mol2')
        self.assertTrue(Mol2File.id_format(self.get_fn('test', written=True)))
        x = Mol2File.parse(self.get_fn('test', written=True))
        self._check_templates(x, self.ace, preserve_headtail=False)
        self.nme.save(self.get_fn('test', written=True), format='mol2')
        self.assertTrue(Mol2File.id_format(self.get_fn('test', written=True)))
        x = Mol2File.parse(self.get_fn('test', written=True))
        self._check_templates(x, self.nme, preserve_headtail=False)
        # Check saving mol2 files by filename extension
        self.ace.save(self.get_fn('test.mol2', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol2', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol2', written=True))
        self._check_templates(x, self.ace, preserve_headtail=False)
        self.nme.save(self.get_fn('test.mol2', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol2', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol2', written=True))
        self._check_templates(x, self.nme, preserve_headtail=False)
        # Now try zipped
        self.ace.save(self.get_fn('test.mol2.gz', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol2.gz', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol2.gz', written=True))
        self._check_templates(x, self.ace, preserve_headtail=False)
        self.nme.save(self.get_fn('test.mol2.gz', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol2.gz', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol2.gz', written=True))
        self._check_templates(x, self.nme, preserve_headtail=False)
        self.ace.save(self.get_fn('test.mol2.bz2', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol2.bz2', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol2.bz2', written=True))
        self._check_templates(x, self.ace, preserve_headtail=False)
        self.nme.save(self.get_fn('test.mol2.bz2', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol2.bz2', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol2.bz2', written=True))
        self._check_templates(x, self.nme, preserve_headtail=False)

    def test_residue_template_pdb_save(self):
        """ Tests ResidueTemplate.save() method for PDB file """
        # Check saving pdb files by keyword
        self.ace.save(self.get_fn('test', written=True), format='pdb')
        self.assertTrue(PDBFile.id_format(self.get_fn('test', written=True)))
        x = PDBFile.parse(self.get_fn('test', written=True))
        self.assertEqual(len(x.atoms), len(self.ace.atoms))
        for a1, a2 in zip(x.atoms, self.ace.atoms):
            self.assertEqual(a1.name, a2.name)
            self.assertEqual(a1.residue.name, a2.residue.name)

        self.nme.save(self.get_fn('test.pdb', written=True))
        self.assertTrue(PDBFile.id_format(self.get_fn('test', written=True)))
        x = PDBFile.parse(self.get_fn('test.pdb', written=True))
        self.assertEqual(len(x.atoms), len(self.nme.atoms))
        for a1, a2 in zip(x.atoms, self.nme.atoms):
            self.assertEqual(a1.name, a2.name)
            self.assertEqual(a1.residue.name, a2.residue.name)

    def test_residue_template_mol3_save(self):
        """ Tests ResidueTemplate.save() method for Mol3 file """
        # Check saving mol3 files by keyword
        self.ace.save(self.get_fn('test', written=True), format='mol3')
        self.assertTrue(Mol2File.id_format(self.get_fn('test', written=True)))
        x = Mol2File.parse(self.get_fn('test', written=True))
        self._check_templates(x, self.ace, preserve_headtail=True)
        self.nme.save(self.get_fn('test', written=True), format='mol3')
        self.assertTrue(Mol2File.id_format(self.get_fn('test', written=True)))
        x = Mol2File.parse(self.get_fn('test', written=True))
        self._check_templates(x, self.nme, preserve_headtail=True)
        # Check saving mol3 files by filename extension
        self.ace.save(self.get_fn('test.mol3', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol3', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol3', written=True))
        self._check_templates(x, self.ace, preserve_headtail=True)
        self.nme.save(self.get_fn('test.mol3', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol3', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol3', written=True))
        self._check_templates(x, self.nme, preserve_headtail=True)
        # Now try zipped
        self.ace.save(self.get_fn('test.mol3.gz', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol3.gz', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol3.gz', written=True))
        self._check_templates(x, self.ace, preserve_headtail=True)
        self.nme.save(self.get_fn('test.mol3.gz', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol3.gz', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol3.gz', written=True))
        self._check_templates(x, self.nme, preserve_headtail=True)
        self.ace.save(self.get_fn('test.mol3.bz2', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol3.bz2', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol3.bz2', written=True))
        self._check_templates(x, self.ace, preserve_headtail=True)
        self.nme.save(self.get_fn('test.mol3.bz2', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol3.bz2', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol3.bz2', written=True))
        self._check_templates(x, self.nme, preserve_headtail=True)

    def test_residue_template_off_save(self):
        """ Tests ResidueTemplate.save() method for OFF lib file """
        # Check saving OFF files by keyword
        self.ace.save(self.get_fn('test', written=True), format='offlib')
        self.assertTrue(AmberOFFLibrary.id_format(self.get_fn('test', written=True)))
        x = AmberOFFLibrary.parse(self.get_fn('test', written=True))['ACE']
        self._check_templates(x, self.ace, preserve_headtail=True)
        self.nme.save(self.get_fn('test', written=True), format='offlib')
        self.assertTrue(AmberOFFLibrary.id_format(self.get_fn('test', written=True)))
        x = AmberOFFLibrary.parse(self.get_fn('test', written=True))['NME']
        self._check_templates(x, self.nme, preserve_headtail=True)
        # Check saving OFF files by filename extension
        self.ace.save(self.get_fn('test.lib', written=True))
        self.assertTrue(AmberOFFLibrary.id_format(self.get_fn('test.lib', written=True)))
        x = AmberOFFLibrary.parse(self.get_fn('test.lib', written=True))['ACE']
        self._check_templates(x, self.ace, preserve_headtail=True)
        self.nme.save(self.get_fn('test.off', written=True))
        self.assertTrue(AmberOFFLibrary.id_format(self.get_fn('test.off', written=True)))
        x = AmberOFFLibrary.parse(self.get_fn('test.off', written=True))['NME']
        self._check_templates(x, self.nme, preserve_headtail=True)
        # Now try zipped
        self.ace.save(self.get_fn('test.lib.gz', written=True))
        self.assertTrue(AmberOFFLibrary.id_format(self.get_fn('test.lib.gz', written=True)))
        x = AmberOFFLibrary.parse(self.get_fn('test.lib.gz', written=True))['ACE']
        self._check_templates(x, self.ace, preserve_headtail=True)
        self.nme.save(self.get_fn('test.off.gz', written=True))
        self.assertTrue(AmberOFFLibrary.id_format(self.get_fn('test.off.gz', written=True)))
        x = AmberOFFLibrary.parse(self.get_fn('test.off.gz', written=True))['NME']
        self._check_templates(x, self.nme, preserve_headtail=True)
        self.ace.save(self.get_fn('test.lib.bz2', written=True))
        self.assertTrue(AmberOFFLibrary.id_format(self.get_fn('test.lib.bz2', written=True)))
        x = AmberOFFLibrary.parse(self.get_fn('test.lib.bz2', written=True))['ACE']
        self._check_templates(x, self.ace, preserve_headtail=True)
        self.nme.save(self.get_fn('test.off.bz2', written=True))
        self.assertTrue(AmberOFFLibrary.id_format(self.get_fn('test.off.bz2', written=True)))
        x = AmberOFFLibrary.parse(self.get_fn('test.off.bz2', written=True))['NME']
        self._check_templates(x, self.nme, preserve_headtail=True)

    def test_residue_template_save_bad_format(self):
        """ Tests proper exceptions for bad format types to ResidueTemplate """
        for bad_fname in ('noextension', 'bad_extension.txt', 'bad_extension.bz2'):
            with self.assertRaises(ValueError):
                self.nme.save(self.get_fn(bad_fname, written=True))
        with self.assertRaises(ValueError):
            self.nme.save(self.get_fn('test.mol2', written=True), format='SOMETHING')

    def test_residue_template_container_mol2_save(self):
        """ Tests ResidueTemplateContainer.save() method for mol2 files """
        # Check saving mol2 files by keyword
        self.container.save(self.get_fn('test', written=True), format='mol2')
        self.assertTrue(Mol2File.id_format(self.get_fn('test', written=True)))
        x = Mol2File.parse(self.get_fn('test', written=True))
        self.assertIsInstance(x, ResidueTemplateContainer)
        self._check_templates(x[0], self.ace, preserve_headtail=False)
        self._check_templates(x[1], self.nme, preserve_headtail=False)
        # Make sure it is a multi-@<MOLECULE> mol2 file (so it can't be loaded
        # as a Structure)
        self.assertRaises(Mol2Error, lambda:
                Mol2File.parse(self.get_fn('test', written=True), structure=True))
        # Check saving mol2 files by filename extension
        self.container.save(self.get_fn('test.mol2', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol2', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol2', written=True))
        self.assertIsInstance(x, ResidueTemplateContainer)
        self._check_templates(x[0], self.ace, preserve_headtail=False)
        self._check_templates(x[1], self.nme, preserve_headtail=False)
        # Now try zipped
        self.container.save(self.get_fn('test.mol2.gz', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol2.gz', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol2.gz', written=True))
        self.assertIsInstance(x, ResidueTemplateContainer)
        self._check_templates(x[0], self.ace, preserve_headtail=False)
        self._check_templates(x[1], self.nme, preserve_headtail=False)
        self.container.save(self.get_fn('test.mol2.bz2', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol2.bz2', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol2.bz2', written=True))
        self.assertIsInstance(x, ResidueTemplateContainer)
        self._check_templates(x[0], self.ace, preserve_headtail=False)
        self._check_templates(x[1], self.nme, preserve_headtail=False)

    def test_residue_template_container_mol3_save(self):
        """ Tests ResidueTemplateContainer.save() method for mol3 files """
        # Check saving mol3 files by keyword
        self.container.save(self.get_fn('test', written=True), format='mol3')
        self.assertTrue(Mol2File.id_format(self.get_fn('test', written=True)))
        x = Mol2File.parse(self.get_fn('test', written=True))
        self.assertIsInstance(x, ResidueTemplateContainer)
        self._check_templates(x[0], self.ace, preserve_headtail=True)
        self._check_templates(x[1], self.nme, preserve_headtail=True)
        # Make sure it is a multi-@<MOLECULE> mol3 file (so it can't be loaded
        # as a Structure)
        with self.assertRaises(Mol2Error):
            Mol2File.parse(self.get_fn('test', written=True), structure=True)
        # Check saving mol3 files by filename extension
        self.container.save(self.get_fn('test.mol3', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol3', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol3', written=True))
        self.assertIsInstance(x, ResidueTemplateContainer)
        self._check_templates(x[0], self.ace, preserve_headtail=True)
        self._check_templates(x[1], self.nme, preserve_headtail=True)
        # Now try zipped
        self.container.save(self.get_fn('test.mol3.gz', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol3.gz', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol3.gz', written=True))
        self.assertIsInstance(x, ResidueTemplateContainer)
        self._check_templates(x[0], self.ace, preserve_headtail=True)
        self._check_templates(x[1], self.nme, preserve_headtail=True)
        self.container.save(self.get_fn('test.mol3.bz2', written=True))
        self.assertTrue(Mol2File.id_format(self.get_fn('test.mol3.bz2', written=True)))
        x = Mol2File.parse(self.get_fn('test.mol3.bz2', written=True))
        self.assertIsInstance(x, ResidueTemplateContainer)
        self._check_templates(x[0], self.ace, preserve_headtail=True)
        self._check_templates(x[1], self.nme, preserve_headtail=True)

    def test_residue_template_container_off_save(self):
        """ Tests ResidueTemplateContainer.save() method for Amber OFF files """
        # Check saving Amber OFF files by keyword
        self.container.save(self.get_fn('test', written=True), format='offlib')
        self.assertTrue(AmberOFFLibrary.id_format(self.get_fn('test', written=True)))
        x = AmberOFFLibrary.parse(self.get_fn('test', written=True))
        self._check_templates(x['ACE'], self.ace, preserve_headtail=True)
        self._check_templates(x['NME'], self.nme, preserve_headtail=True)
        # Check saving mol2 files by filename extension
        self.container.save(self.get_fn('test.off', written=True))
        self.assertTrue(AmberOFFLibrary.id_format(self.get_fn('test.off', written=True)))
        x = AmberOFFLibrary.parse(self.get_fn('test.off', written=True))
        self._check_templates(x['ACE'], self.ace, preserve_headtail=True)
        self._check_templates(x['NME'], self.nme, preserve_headtail=True)
        # Now try zipped
        self.container.save(self.get_fn('test.lib.gz', written=True))
        self.assertTrue(AmberOFFLibrary.id_format(self.get_fn('test.lib.gz', written=True)))
        x = AmberOFFLibrary.parse(self.get_fn('test.lib.gz', written=True))
        self._check_templates(x['ACE'], self.ace, preserve_headtail=True)
        self._check_templates(x['NME'], self.nme, preserve_headtail=True)
        self.container.save(self.get_fn('test.lib.bz2', written=True))
        self.assertTrue(AmberOFFLibrary.id_format(self.get_fn('test.lib.bz2', written=True)))
        x = AmberOFFLibrary.parse(self.get_fn('test.lib.bz2', written=True))
        self._check_templates(x['ACE'], self.ace, preserve_headtail=True)
        self._check_templates(x['NME'], self.nme, preserve_headtail=True)

    def test_bad_save(self):
        """ Test error handling in ResidueTemplateContainer.save """
        self.assertRaises(ValueError, lambda:
                self.container.save('unrecognized.file.bz2'))
        self.assertRaises(ValueError, lambda:
                self.container.save('unrecognized.file'))
        self.assertRaises(ValueError, lambda:
                self.container.save('something', format='NOPE'))

    def _check_templates(self, templ1, templ2, preserve_headtail=True):
        self.assertEqual(len(templ1.atoms), len(templ2.atoms))
        self.assertEqual(len(templ1.bonds), len(templ2.bonds))
        if preserve_headtail:
            if templ1.head is None or templ2.head is None:
                self.assertIs(templ1.head, None)
                self.assertIs(templ2.head, None)
            else:
                self.assertIsNot(templ1.head, None)
                self.assertIsNot(templ2.head, None)
                self.assertEqual(templ1.head.idx, templ2.head.idx)
            if templ1.tail is None or templ2.tail is None:
                self.assertIs(templ1.tail, None)
                self.assertIs(templ2.tail, None)
            else:
                self.assertIsNot(templ1.tail, None)
                self.assertIsNot(templ2.tail, None)
                self.assertEqual(templ1.tail.idx, templ2.tail.idx)
        elif preserve_headtail is not None:
            self.assertIs(templ1.head, None)
            self.assertIs(templ1.tail, None)
        bs1 = set()
        bs2 = set()
        for b1, b2 in zip(templ1.bonds, templ2.bonds):
            bs1.add(tuple(sorted([b1.atom1.idx, b1.atom2.idx])))
            bs2.add(tuple(sorted([b2.atom1.idx, b2.atom2.idx])))
        self.assertEqual(bs1, bs2)

class TestAmberOFFLibrary(utils.FileIOTestCase):
    """ Tests the AmberOFFLibrary class """

    def test_from_library(self):
        """ Tests ResidueTemplateContainer.from_library functionality """
        offlib = AmberOFFLibrary.parse(get_fn('amino12.lib'))
        lib = ResidueTemplateContainer.from_library(offlib)
        self.assertIsInstance(lib, ResidueTemplateContainer)
        self.assertEqual(len(lib), len(offlib))
        for r1, r2 in zip(offlib, lib):
            r1 = offlib[r1]
            self.assertIs(r1, r2)
        lib2 = ResidueTemplateContainer.from_library(offlib, copy=True)
        for r1, r2 in zip(offlib, lib2):
            r1 = offlib[r1]
            self.assertIsNot(r1, r2)

    def test_read_internal(self):
        """ Tests reading Amber amino12 OFF library (internal residues) """
        offlib = AmberOFFLibrary.parse(get_fn('amino12.lib'))
        self.assertEqual(len(offlib), 28)
        for name, res in offlib.items():
            self.assertIsInstance(res, ResidueTemplate)
            self.assertEqual(name, res.name)
            self.assertEqual(res.head.name, 'N')
            self.assertEqual(res.tail.name, 'C')
            self.assertIs(res.type, PROTEIN)
        # Check two residues in particular: ALA and CYX
        ala = offlib['ALA']
        self.assertEqual(len(ala), 10)
        self.assertEqual(len(ala.bonds), 9)
        self.assertIn(ala[0], ala[1].bond_partners)
        self.assertIn(ala[0], ala[2].bond_partners)
        self.assertIn(ala[2], ala[3].bond_partners)
        self.assertIn(ala[2], ala[4].bond_partners)
        self.assertIn(ala[2], ala[8].bond_partners)
        self.assertIn(ala[4], ala[5].bond_partners)
        self.assertIn(ala[4], ala[6].bond_partners)
        self.assertIn(ala[4], ala[7].bond_partners)
        self.assertIn(ala[8], ala[9].bond_partners)
        self.assertAlmostEqual(ala[0].xx, 3.325770)
        self.assertAlmostEqual(ala[0].xy, 1.547909)
        self.assertAlmostEqual(ala[0].xz, -1.607204E-06)
        self.assertAlmostEqual(ala[1].xx, 3.909407)
        self.assertAlmostEqual(ala[1].xy, 0.723611)
        self.assertAlmostEqual(ala[1].xz, -2.739882E-06)
        self.assertAlmostEqual(ala[2].xx, 3.970048)
        self.assertAlmostEqual(ala[2].xy, 2.845795)
        self.assertAlmostEqual(ala[2].xz, -1.311163E-07)
        self.assertAlmostEqual(ala[3].xx, 3.671663)
        self.assertAlmostEqual(ala[3].xy, 3.400129)
        self.assertAlmostEqual(ala[3].xz, -0.889820)
        self.assertAlmostEqual(ala[4].xx, 3.576965)
        self.assertAlmostEqual(ala[4].xy, 3.653838)
        self.assertAlmostEqual(ala[4].xz, 1.232143)
        self.assertAlmostEqual(ala[5].xx, 3.877484)
        self.assertAlmostEqual(ala[5].xy, 3.115795)
        self.assertAlmostEqual(ala[5].xz, 2.131197)
        self.assertAlmostEqual(ala[6].xx, 4.075059)
        self.assertAlmostEqual(ala[6].xy, 4.623017)
        self.assertAlmostEqual(ala[6].xz, 1.205786)
        self.assertAlmostEqual(ala[7].xx, 2.496995)
        self.assertAlmostEqual(ala[7].xy, 3.801075)
        self.assertAlmostEqual(ala[7].xz, 1.241379)
        self.assertAlmostEqual(ala[8].xx, 5.485541)
        self.assertAlmostEqual(ala[8].xy, 2.705207)
        self.assertAlmostEqual(ala[8].xz, -4.398755E-06)
        self.assertAlmostEqual(ala[9].xx, 6.008824)
        self.assertAlmostEqual(ala[9].xy, 1.593175)
        self.assertAlmostEqual(ala[9].xz, -8.449768E-06)
        # now cyx
        cyx = offlib['CYX']
        self.assertEqual(len(cyx), 10)
        self.assertEqual(len(cyx.bonds), 9)
        self.assertIn(cyx[0], cyx[1].bond_partners)
        self.assertIn(cyx[0], cyx[2].bond_partners)
        self.assertIn(cyx[2], cyx[3].bond_partners)
        self.assertIn(cyx[2], cyx[4].bond_partners)
        self.assertIn(cyx[2], cyx[8].bond_partners)
        self.assertIn(cyx[4], cyx[5].bond_partners)
        self.assertIn(cyx[4], cyx[6].bond_partners)
        self.assertIn(cyx[4], cyx[7].bond_partners)
        self.assertIn(cyx[8], cyx[9].bond_partners)
        self.assertAlmostEqual(cyx[0].xx, 3.325770)
        self.assertAlmostEqual(cyx[0].xy, 1.547909)
        self.assertAlmostEqual(cyx[0].xz, -1.607204E-06)
        self.assertAlmostEqual(cyx[1].xx, 3.909407)
        self.assertAlmostEqual(cyx[1].xy, 0.723611)
        self.assertAlmostEqual(cyx[1].xz, -2.739882E-06)
        self.assertAlmostEqual(cyx[2].xx, 3.970048)
        self.assertAlmostEqual(cyx[2].xy, 2.845795)
        self.assertAlmostEqual(cyx[2].xz, -1.311163E-07)
        self.assertAlmostEqual(cyx[3].xx, 3.671663)
        self.assertAlmostEqual(cyx[3].xy, 3.400129)
        self.assertAlmostEqual(cyx[3].xz, -0.889820)
        self.assertAlmostEqual(cyx[4].xx, 3.576965)
        self.assertAlmostEqual(cyx[4].xy, 3.653838)
        self.assertAlmostEqual(cyx[4].xz, 1.232143)
        self.assertAlmostEqual(cyx[5].xx, 2.496995)
        self.assertAlmostEqual(cyx[5].xy, 3.801075)
        self.assertAlmostEqual(cyx[5].xz, 1.241379)
        self.assertAlmostEqual(cyx[6].xx, 3.877484)
        self.assertAlmostEqual(cyx[6].xy, 3.115795)
        self.assertAlmostEqual(cyx[6].xz, 2.131197)
        self.assertAlmostEqual(cyx[7].xx, 4.309573)
        self.assertAlmostEqual(cyx[7].xy, 5.303523)
        self.assertAlmostEqual(cyx[7].xz, 1.366036)
        self.assertAlmostEqual(cyx[8].xx, 5.485541)
        self.assertAlmostEqual(cyx[8].xy, 2.705207)
        self.assertAlmostEqual(cyx[8].xz, -4.398755E-06)
        self.assertAlmostEqual(cyx[9].xx, 6.008824)
        self.assertAlmostEqual(cyx[9].xy, 1.593175)
        self.assertAlmostEqual(cyx[9].xz, -8.449768E-06)
        # Check connections
        self.assertEqual(len(cyx.connections), 1)
        self.assertEqual(cyx.connections[0].name, 'SG')

    def test_read_n_term(self):
        """ Test reading N-terminal amino acid Amber OFF library """
        offlib = AmberOFFLibrary.parse(get_fn('aminont12.lib'))
        self.assertEqual(len(offlib), 24)
        for name, res in offlib.items():
            self.assertIsInstance(res, ResidueTemplate)
            self.assertEqual(name, res.name)
            self.assertIs(res.head, None)
            self.assertEqual(res.tail.name, 'C')
            self.assertIs(res.type, PROTEIN)

    def test_read_c_term(self):
        """ Test reading C-terminal amino acid Amber OFF library """
        offlib = AmberOFFLibrary.parse(get_fn('aminoct12.lib'))
        self.assertEqual(len(offlib), 26)
        for name, res in offlib.items():
            self.assertIsInstance(res, ResidueTemplate)
            self.assertEqual(name, res.name)
            self.assertEqual(res.head.name, 'N')
            self.assertIs(res.type, PROTEIN)

    def test_read_solvents(self):
        """ Test reading solvent Amber OFF lib (multi-res units) """
        # Turn off warnings... the solvents.lib file is SO broken.
        offlib = AmberOFFLibrary.parse(get_fn('solvents.lib'))
        self.assertEqual(len(offlib), 24)
        for name, res in offlib.items():
            self.assertEqual(res.name, name)
            if 'BOX' in name:
                self.assertIsInstance(res, ResidueTemplateContainer)
                # Make sure all residues have the same features as the first
                for r in res:
                    self.assertIs(r.type, SOLVENT)
                    for a1, a2 in zip(r, res[0]):
                        self.assertEqual(a1.name, a2.name)
                        self.assertEqual(a1.type, a2.type)
                        self.assertEqual(a1.charge, a2.charge)
                        self.assertEqual(a1.atomic_number, a2.atomic_number)
                        self.assertEqual(len(a1.bond_partners),
                                         len(a2.bond_partners))
                        set1 = set([x.name for x in a1.bond_partners])
                        set2 = set([x.name for x in a2.bond_partners])
                        self.assertEqual(set1, set2)
                        if a1 is not a2:
                            self.assertNotEqual(a1.xx, a2.xx)
                            self.assertNotEqual(a1.xy, a2.xy)
                            self.assertNotEqual(a1.xz, a2.xz)
            else:
                self.assertIs(res.type, SOLVENT)
                self.assertIsInstance(res, ResidueTemplate)
        # Check a few solvent boxes in particular
        chcl3 = offlib['CHCL3BOX']
        self.assertEqual(chcl3.name, 'CHCL3BOX')
        self.assertEqual(len(chcl3), 1375)
        self.assertEqual(chcl3.box[0], 56.496)
        self.assertEqual(chcl3.box[1], 56.496)
        self.assertEqual(chcl3.box[2], 56.496)
        for res in chcl3:
            self.assertEqual(res.name, 'CL3')
        self.assertAlmostEqual(chcl3.box[3], 90, places=4)
        self.assertAlmostEqual(chcl3.box[4], 90, places=4)
        self.assertAlmostEqual(chcl3.box[5], 90, places=4)
        # Check some positions (but obviously not all)
        self.assertAlmostEqual(chcl3[0][0].xx, -22.675111)
        self.assertAlmostEqual(chcl3[0][0].xy, -13.977137)
        self.assertAlmostEqual(chcl3[0][0].xz, -21.470579)
        self.assertAlmostEqual(chcl3[1][0].xx, -9.668111)
        self.assertAlmostEqual(chcl3[1][0].xy, -15.097137)
        self.assertAlmostEqual(chcl3[1][0].xz, -18.569579)
        # We can't convert the solvents library to a ResidueTemplateContainer,
        # since it contains ResidueTemplateContainer instances
        self.assertRaises(ValueError, lambda:
                ResidueTemplateContainer.from_library(offlib))

    def test_read_write_internal(self):
        """ Tests reading/writing of Amber OFF internal AA libs """
        offlib = AmberOFFLibrary.parse(get_fn('amino12.lib'))
        outfile = StringIO()
        AmberOFFLibrary.write(offlib, outfile)
        outfile.seek(0)
        offlib2 = AmberOFFLibrary.parse(outfile)
        self._check_read_written_libs(offlib, offlib2)

    def test_read_write_c_term(self):
        """ Tests reading/writing of Amber OFF C-terminal AA libs """
        offlib = AmberOFFLibrary.parse(get_fn('aminoct12.lib'))
        outfile = StringIO()
        AmberOFFLibrary.write(offlib, outfile)
        outfile.seek(0)
        offlib2 = AmberOFFLibrary.parse(outfile)
        self._check_read_written_libs(offlib, offlib2)

    def test_read_write_n_term(self):
        """ Tests reading/writing of Amber OFF N-terminal AA libs """
        offlib = AmberOFFLibrary.parse(get_fn('aminont12.lib'))
        outfile = StringIO()
        AmberOFFLibrary.write(offlib, outfile)
        outfile.seek(0)
        offlib2 = AmberOFFLibrary.parse(outfile)
        self._check_read_written_libs(offlib, offlib2)

    def test_read_write_solvent_lib(self):
        """ Tests reading/writing of Amber OFF solvent libs """
        offlib = AmberOFFLibrary.parse(get_fn('solvents.lib'))
        outfile = StringIO()
        AmberOFFLibrary.write(offlib, outfile)
        outfile.seek(0)
        offlib2 = AmberOFFLibrary.parse(outfile)

    def test_bad_off_files(self):
        """ Tests error checking in OFF library files """
        self.assertRaises(ValueError, lambda:
                AmberOFFLibrary.parse(get_fn('trx.prmtop')))
        with open(self.get_fn('test.off', written=True), 'w') as f:
            with open(get_fn('amino12.lib'), 'r') as ff:
                for i in range(10):
                    f.write(ff.readline())
        with self.assertRaises(RuntimeError):
            AmberOFFLibrary.parse(self.get_fn('test.off', written=True))

    @unittest.skipIf(pd is None, "Cannot test without pandas")
    def test_data_frame(self):
        """ Test converting ResidueTemplate to a DataFrame """
        offlib = AmberOFFLibrary.parse(get_fn('amino12.lib'))
        df = offlib['ALA'].to_dataframe()
        self.assertEqual(df.shape, (10, 26))
        self.assertAlmostEqual(df.charge.sum(), 0)
        self.assertEqual(df.atomic_number.sum(), 38)

    def _check_read_written_libs(self, offlib, offlib2):
        # Check that offlib and offlib2 are equivalent
        self.assertEqual(len(offlib), len(offlib2))
        self.assertEqual(offlib.keys(), offlib2.keys())
        for key in offlib.keys():
            r1 = offlib[key]
            r2 = offlib2[key]
            # Check residues
            self.assertEqual(len(r1), len(r2))
            self.assertIs(r1.type, r2.type)
            # Check head and tail
            if r1.head is None or r2.head is None:
                self.assertIs(r1.head, None)
                self.assertIs(r2.head, None)
            else:
                self.assertEqual(r1.head.name, r2.head.name)
                self.assertEqual(r1.head.type, r2.head.type)
                self.assertEqual(r1.head.idx, r2.head.idx)
            if r1.tail is None or r2.tail is None:
                self.assertIs(r1.tail, None)
                self.assertIs(r2.tail, None)
            else:
                self.assertEqual(r1.tail.name, r2.tail.name)
                self.assertEqual(r1.tail.type, r2.tail.type)
                self.assertEqual(r1.tail.idx, r2.tail.idx)
            # Check atom properties
            for a1, a2 in zip(r1, r2):
                self.assertEqual(a1.name, a2.name)
                self.assertEqual(a1.type, a2.type)
                self.assertAlmostEqual(a1.charge, a2.charge)
                self.assertAlmostEqual(a1.xx, a2.xx, places=4)
                self.assertAlmostEqual(a1.xy, a2.xy, places=4)
                self.assertAlmostEqual(a1.xz, a2.xz, places=4)
                self.assertEqual(a1.vx, a2.vx)
                self.assertEqual(a1.vy, a2.vy)
                self.assertEqual(a1.vz, a2.vz)
            # Check bonds
            self.assertEqual(len(r1.bonds), len(r2.bonds))
            for b1, b2 in zip(r1.bonds, r2.bonds):
                self.assertEqual(b1.atom1.name, b2.atom1.name)
                self.assertEqual(b1.atom2.name, b2.atom2.name)

class TestAmberOFFLeapCompatibility(utils.FileIOTestCase):
    """ Tests the AmberOFFLibrary classes written in LEaP """

    def setUp(self):
        super().setUp()
        self.tleap = utils.which('tleap')
        self.cwd = os.getcwd()
        os.chdir(self._temporary_directory.name)

    def tearDown(self):
        os.chdir(self.cwd)
        super().tearDown()

    @unittest.skipIf(utils.which('tleap') is None, "Cannot test without tleap")
    def test_amber_amino_internal(self):
        """ Test that the internal AA OFF library writes work with LEaP """
        # First create the parm to test against... we are in "writes" right now
        offlib = AmberOFFLibrary.parse(get_fn('amino12.lib'))
        AmberOFFLibrary.write(offlib, 'testinternal.lib')
        f = open('tleap_orig.in', 'w')
        f.write("""\
source "%s"
l = sequence {ALA ARG ASH ASN ASP CYM CYS CYX GLH GLN GLU GLY HID HIE HIP \
              HYP ILE LEU LYN LYS MET PHE PRO SER THR TRP TYR VAL}
set default PBRadii mbondi2
savePDB l alphabet.pdb
saveAmberParm l alphabet.parm7 alphabet.rst7
quit
""" % get_fn('leaprc.ff12SB'))
        f.close()
        # Now create the leaprc for our new files
        f = open('tleap_new.in', 'w')
        f.write("""\
loadAmberParams parm10.dat
loadAmberParams frcmod.ff12SB
loadOFF testinternal.lib
l = sequence {ALA ARG ASH ASN ASP CYM CYS CYX GLH GLN GLU GLY HID HIE HIP \
              HYP ILE LEU LYN LYS MET PHE PRO SER THR TRP TYR VAL}
savePDB l alphabet2.pdb
saveAmberParm l alphabet2.parm7 alphabet2.rst7
quit
""")
        f.close()
        os.system('tleap -f tleap_orig.in > tleap_orig.out 2>&1')
        os.system('tleap -f tleap_new.in > tleap_new.out 2>&1')
        # Compare the resulting files
        pdb1 = read_PDB('alphabet.pdb')
        pdb2 = read_PDB('alphabet2.pdb')
        parm1 = AmberParm('alphabet.parm7', 'alphabet.rst7')
        parm2 = AmberParm('alphabet2.parm7', 'alphabet2.rst7')
        # Since there are some specific parts of the leaprc that affect default
        # radii, change it here intentionally
        changeRadii(parm1, 'mbondi2').execute()
        changeRadii(parm2, 'mbondi2').execute()
        self._check_corresponding_files(pdb1, pdb2, parm1, parm2)

    @unittest.skipIf(utils.which('tleap') is None, "Cannot test without tleap")
    @pytest.mark.xfail
    def test_amber_amino_termini(self):
        """ Test that the terminal AA OFF library writes work with LEaP """
        offlib_nter = AmberOFFLibrary.parse(get_fn('aminont12.lib'))
        offlib_cter = AmberOFFLibrary.parse(get_fn('aminoct12.lib'))
        AmberOFFLibrary.write(offlib_nter, 'testnt.lib')
        AmberOFFLibrary.write(offlib_cter, 'testct.lib')
        # Test all pairs a random set of 10 pairs
        keys1 = [random.choice(list(offlib_nter.keys())) for i in range(10)]
        keys2 = [random.choice(list(offlib_cter.keys())) for i in range(10)]
        for key1, key2 in zip(keys1, keys2):
            f = open('tleap_orig.in', 'w')
            f.write("""\
source "%s"
l = sequence {%s %s}
savePDB l alphabet.pdb
saveAmberParm l alphabet.parm7 alphabet.rst7
quit
""" % (get_fn('leaprc.ff12SB'), key1, key2))
            f.close()
            f = open('tleap_new.in', 'w')
            f.write("""\
loadAmberParams parm10.dat
loadAmberParams frcmod.ff12SB
loadOFF testct.lib
loadOFF testnt.lib
l = sequence {%s %s}
savePDB l alphabet2.pdb
saveAmberParm l alphabet2.parm7 alphabet2.rst7
quit
""" % (key1, key2))
            f.close()
            os.system('tleap -f tleap_orig.in > tleap_orig.out 2>&1')
            os.system('tleap -f tleap_new.in > tleap_new.out 2>&1')
            # Compare the resulting files
            pdb1 = read_PDB('alphabet.pdb')
            pdb2 = read_PDB('alphabet2.pdb')
            parm1 = AmberParm('alphabet.parm7', 'alphabet.rst7')
            parm2 = AmberParm('alphabet2.parm7', 'alphabet2.rst7')
            # Since there are some specific parts of the leaprc that affect
            # default radii, change it here intentionally
            changeRadii(parm1, 'mbondi2').execute()
            changeRadii(parm2, 'mbondi2').execute()
            self._check_corresponding_files(pdb1, pdb2, parm1, parm2, False)

    def _check_corresponding_files(self, pdb1, pdb2, parm1, parm2, tree=True):
        self.assertEqual(len(pdb1.atoms), len(pdb2.atoms))
        self.assertEqual(len(parm1.atoms), len(parm2.atoms))
        self.assertEqual(len(parm1.bonds), len(parm2.bonds))
        for a1, a2 in zip(pdb1.atoms, pdb2.atoms):
            self.assertEqual(a1.name, a2.name, 'pdb1 atoms:\n{}\npdb2 atoms:\n{}\n'.format(pdb1.atoms, pdb2.atoms))
            self.assertEqual(a1.atomic_number, a2.atomic_number)
        for a1, a2 in zip(parm1.atoms, parm2.atoms):
            # Check EVERYTHING
            self.assertIsNot(a1, a2)
            self.assertEqual(a1.name, a2.name)
            self.assertEqual(a1.type, a2.type)
            self.assertEqual(a1.nb_idx, a2.nb_idx)
            self.assertEqual(a1.atomic_number, a2.atomic_number)
            self.assertEqual(a1.atom_type.rmin, a2.atom_type.rmin)
            self.assertEqual(a1.atom_type.epsilon, a2.atom_type.epsilon)
            self.assertEqual(a1.solvent_radius, a2.solvent_radius)
            self.assertEqual(a1.screen, a2.screen)
            # Ugh. OFF libs are inconsistent
            if tree:
                self.assertEqual(a1.tree, a2.tree)
            self.assertEqual(len(a1.bonds), len(a2.bonds))
            self.assertEqual(len(a1.angles), len(a2.angles))
            self.assertEqual(len(a1.dihedrals), len(a2.dihedrals))
            set1 = set([a.name for a in a1.bond_partners])
            set2 = set([a.name for a in a2.bond_partners])
            self.assertEqual(set1, set2)
            set1 = set([a.name for a in a1.angle_partners])
            set2 = set([a.name for a in a2.angle_partners])
            self.assertEqual(set1, set2)
            set1 = set([a.name for a in a1.dihedral_partners])
            set2 = set([a.name for a in a2.dihedral_partners])
            self.assertEqual(set1, set2)
            # Check residue properties
            self.assertEqual(a1.residue.name, a2.residue.name)

class TestSlice(unittest.TestCase):
    '''Test slicing ResidueTemplate'''

    def test_slice_from_array_like(self):
        """ Test slicing by a tuple/list"""
        residue = pmd.load_file(get_fn('aminont12.lib'))['NALA']

        names = ['CA', 'CB', 'C', 'N']
        for op in [list, tuple]:
            atomlist = residue[op(names)]
            self.assertEqual(len(names), len(atomlist))
            for atom in atomlist:
                self.assertEqual(atom.name, residue[atom.name].name)

        indices = [0, 4, 7, 3]
        for op in [list, tuple]:
            atomlist = residue[op(indices)]
            self.assertEqual(len(indices), len(atomlist))
            for atom in atomlist:
                self.assertEqual(atom.name, residue[atom.name].name)

class TestBondDetermination(utils.FileIOTestCase):
    """ Tests assigning bonds to structures """

    def test_standard_residue_database(self):
        """ Tests the availability of standard residue templates """
        for resname in ('ALA', 'ARG', 'ASN', 'ASP', 'CYS', 'GLU', 'GLN', 'GLY',
                        'HIS', 'HYP', 'ILE', 'LEU', 'LYS', 'MET', 'PHE', 'PRO',
                        'SER', 'THR', 'TRP', 'TYR', 'VAL', 'DA', 'DT', 'DG',
                        'DC', 'A', 'U', 'G', 'C', 'ACE', 'NME'):
            self.assertIn(resname, StandardBiomolecularResidues)
            self.assertIsInstance(StandardBiomolecularResidues[resname],
                                  ResidueTemplate)

    def test_simple_bond_assignment(self):
        """ Tests the assignment of bonds to simple Structure instances """
        for name, res in StandardBiomolecularResidues.items():
            s = Structure()
            for a in res.atoms:
                s.add_atom(copy(a), name, 1)
            s.assign_bonds()
            # Now make sure we have the same bond
            self.assertEqual(len(res.atoms), len(s.atoms))
            self.assertEqual(len(res.bonds), len(s.bonds))
            for sa, ra in zip(res.atoms, s.atoms):
                self.assertEqual(sa.name, ra.name)
                self.assertEqual(len(sa.bond_partners), len(ra.bond_partners))
                self.assertEqual({a.name for a in sa.bond_partners},
                                 {a.name for a in ra.bond_partners})

    def test_headtail_bond_assignment(self):
        """ Tests assignment of bonds to polymeric Structures """
        s = read_PDB(get_fn('ava.pdb'))
        # Make sure PDB files have their bond automatically determined
        self.assertGreater(len(s.bonds), 0)
        self.assertIn(s.view[0, 'C'].atoms[0], s.view[1, 'N'].atoms[0].bond_partners)
        self.assertIn(s.view[1, 'C'].atoms[0], s.view[2, 'N'].atoms[0].bond_partners)
        self.assertIn(s.view[2, 'C'].atoms[0], s.view[3, 'N'].atoms[0].bond_partners)
        self.assertIn(s.view[3, 'C'].atoms[0], s.view[4, 'N'].atoms[0].bond_partners)
        # TER card -- nobody bonded to *next* residue
        for a in s.residues[4].atoms:
            for ba in a.bond_partners:
                self.assertNotEqual(ba.residue.idx, 5)
        self.assertIn(s.view[5, 'C'].atoms[0], s.view[6, 'N'].atoms[0].bond_partners)
        self.assertIn(s.view[6, 'C'].atoms[0], s.view[7, 'N'].atoms[0].bond_partners)
        self.assertIn(s.view[7, 'C'].atoms[0], s.view[8, 'N'].atoms[0].bond_partners)
        self.assertIn(s.view[8, 'C'].atoms[0], s.view[9, 'N'].atoms[0].bond_partners)
        # Make sure we have exactly 86 bonds defined
        self.assertEqual(len(s.bonds), 86)

    def test_connect_parsing(self):
        """ Tests processing of PDB CONECT records and see that it adds bonds """
        s = read_PDB(get_fn('4lzt.pdb'))
        # Check that the CONECT record bond specs are actual bonds
        self.assertIn(s.view[5, 'SG'].atoms[0], s.view[126, 'SG'].atoms[0].bond_partners)

    def test_bond_distance_assignment(self):
        """ Tests assignment of disulfides if no CONECT records available """
        fn = self.get_fn('test.pdb', written=True)
        with open(get_fn('4lzt.pdb'), 'r') as f, open(fn, 'w') as fw:
            for line in f:
                if line.startswith('CONECT'): continue
                fw.write(line)
        s = read_PDB(fn)
        # Check that the disulfide is present even without CONECT records
        self.assertIn(s.view[5, 'SG'].atoms[0], s.view[126, 'SG'].atoms[0].bond_partners)

    def test_pairlist(self):
        """ Tests pairlist builder """
        s = Structure()
        for i in range(2000):
            s.add_atom(Atom('XYZ'), 'RES', i)
        s.coordinates = np.random.rand(2000, 3) * 20 - 10
        pairs = find_atom_pairs(s, 5.0)
        self.assertTrue(any(len(x) > 0 for x in pairs))
        for a1 in s.atoms:
            for a2 in s.atoms:
                if a1 is a2: continue
                if distance2(a1, a2) < 25:
                    self.assertIn(a1, pairs[a2.idx])
                    self.assertIn(a2, pairs[a1.idx])

    def test_element_override(self):
        """ Tests that templates improve element information """
        f = StringIO("""\
ATOM      1  CA   CA A   1      -0.641  26.272   5.586  1.00 24.68
""")
        s = read_PDB(f)
        self.assertEqual(s[0].atomic_number, pmd.periodic_table.AtomicNum['Ca'])



"""
Tests the functionality in parmed.residue
"""
import utils
import parmed as pmd
from parmed import residue
from parmed.residue import RNAResidue, DNAResidue
import unittest

class TestChemistryResidue(unittest.TestCase):

    def test_residue_members(self):
        """ Test that the main amino acid residues are members of residue """
        self.assertTrue(hasattr(residue, 'ASP'))
        self.assertTrue(hasattr(residue, 'GLU'))
        self.assertTrue(hasattr(residue, 'HIS'))
        self.assertTrue(hasattr(residue, 'GLN'))
        self.assertTrue(hasattr(residue, 'ASN'))
        self.assertTrue(hasattr(residue, 'LYS'))
        self.assertTrue(hasattr(residue, 'TYR'))
        self.assertTrue(hasattr(residue, 'TRP'))
        self.assertTrue(hasattr(residue, 'ILE'))
        self.assertTrue(hasattr(residue, 'LEU'))
        self.assertTrue(hasattr(residue, 'ALA'))
        self.assertTrue(hasattr(residue, 'GLY'))
        self.assertTrue(hasattr(residue, 'PRO'))
        self.assertTrue(hasattr(residue, 'VAL'))
        self.assertTrue(hasattr(residue, 'THR'))
        self.assertTrue(hasattr(residue, 'SER'))
        self.assertTrue(hasattr(residue, 'CYS'))
        self.assertTrue(hasattr(residue, 'MET'))
        self.assertTrue(hasattr(residue, 'ARG'))
        self.assertTrue(hasattr(residue, 'PHE'))

    def test_name_lookup(self):
        """ Test that looking up residues by name works """
        self.assertIs(residue.AminoAcidResidue.get('Alanine'), residue.ALA)
        self.assertIs(residue.AminoAcidResidue.get('Glutamine'), residue.GLN)
        self.assertIs(residue.AminoAcidResidue.get('glutamate'), residue.GLU)
        self.assertIs(residue.AminoAcidResidue.get('HiStIDiNe'), residue.HIS)
        self.assertIs(residue.AminoAcidResidue.get('ISOLEUCINE'), residue.ILE)

    def test_abbr_lookup(self):
        """ Test that looking up residues by abbreviation works """
        self.assertIs(residue.AminoAcidResidue.get('ALA'), residue.ALA)
        self.assertIs(residue.AminoAcidResidue.get('GLY'), residue.GLY)
        self.assertIs(residue.AminoAcidResidue.get('glu'), residue.GLU)
        self.assertIs(residue.AminoAcidResidue.get('Asp'), residue.ASP)

    def test_symbol_lookup(self):
        """ Test that looking up residues by symbol works """
        self.assertIs(residue.AminoAcidResidue.get('E'), residue.GLU)
        self.assertIs(residue.AminoAcidResidue.get('D'), residue.ASP)
        self.assertIs(residue.AminoAcidResidue.get('A'), residue.ALA)

    def test_repr_output(self):
        """ Test the %r representation of the Amino Acids """
        for res in residue.AminoAcidResidue.all_residues:
            self.assertEqual('<Amino Acid Residue %s: %s [%s]>' % (res.name,
                res.abbr, res.symbol), repr(res))

    def test_bad_lookup(self):
        """ Test that lookups of non-existent residues fails """
        self.assertRaises(KeyError,
                lambda: residue.AminoAcidResidue.get('NoResidue'))

    def test_has(self):
        """ Tests the `has` method of BiomolecularResidue """
        self.assertTrue(residue.AminoAcidResidue.has('E'))
        self.assertTrue(residue.AminoAcidResidue.has('GLU'))
        self.assertTrue(residue.AminoAcidResidue.has(residue.GLU))
        # Now test some that should be False
        self.assertFalse(residue.AminoAcidResidue.has(residue.A))
        self.assertFalse(residue.AminoAcidResidue.has('GUA'))
        self.assertFalse(residue.AminoAcidResidue.has('DG'))

    def test_aliases(self):
        """ Tests that the common aliases used by Amber works """
        self.assertTrue(residue.AminoAcidResidue.has('GL4'))
        self.assertTrue(residue.AminoAcidResidue.has('GLH'))
        self.assertTrue(residue.AminoAcidResidue.has('ASH'))
        self.assertIs(residue.AminoAcidResidue.get('ASH'), residue.ASP)
        self.assertIs(residue.AminoAcidResidue.get('GL4'), residue.GLU)

    def test_termini(self):
        """ Tests that decorated termini names are properly recognized """
        self.assertTrue(residue.AminoAcidResidue.has('CGLY'))
        self.assertIs(residue.AminoAcidResidue.get('CHIS'), residue.HIS)
        self.assertIs(residue.AminoAcidResidue.get('CASH'), residue.ASP)
        self.assertIs(residue.AminoAcidResidue.get('NHIS'), residue.HIS)
        self.assertIs(residue.AminoAcidResidue.get('NASH'), residue.ASP)
        with self.assertRaises(KeyError):
            residue.AminoAcidResidue.get('XASH')

class TestNucleicAcidResidues(unittest.TestCase):

    def test_residue_members(self):
        """ Tests that all of the nucleic acid residues are defined """
        self.assertTrue(hasattr(residue, 'A'))
        self.assertTrue(hasattr(residue, 'U'))
        self.assertTrue(hasattr(residue, 'G'))
        self.assertTrue(hasattr(residue, 'C'))
        self.assertTrue(hasattr(residue, 'DA'))
        self.assertTrue(hasattr(residue, 'DT'))
        self.assertTrue(hasattr(residue, 'DG'))
        self.assertTrue(hasattr(residue, 'DC'))

    def test_name_lookup(self):
        """ Test that looking up DNA/RNA residues by name works """
        self.assertIs(residue.DNAResidue.get('Guanine'), residue.DG)
        self.assertIs(residue.DNAResidue.get('ADE'), residue.DA)
        self.assertIs(residue.DNAResidue.get('thymine'), residue.DT)
        self.assertIs(residue.RNAResidue.get('Uracil'), residue.U)
        self.assertIs(residue.RNAResidue.get('G'), residue.G)
        self.assertIs(residue.RNAResidue.get('ADE'), residue.A)

    def test_bad_lookup(self):
        """ Test that lookups of non-existent nucleic acid residues fails """
        self.assertRaises(KeyError, lambda: residue.RNAResidue.get('NoResidue'))
        self.assertRaises(KeyError, lambda: residue.DNAResidue.get('NoResidue'))

    def test_termini(self):
        """ Tests that decorated DNA/RNA termini are properly recognized """
        self.assertTrue(residue.RNAResidue.has('G5'))
        self.assertTrue(residue.RNAResidue.has('G3'))
        self.assertTrue(residue.RNAResidue.has('RG5'))
        self.assertTrue(residue.RNAResidue.has('RG3'))
        self.assertTrue(residue.DNAResidue.has('DG5'))
        self.assertTrue(residue.DNAResidue.has('DG3'))
        self.assertFalse(residue.RNAResidue.has('G4'))



""" Package that provides an API to Amber-specific files """

__author__ = "Jason Swails <jason.swails@gmail.com>"

__all__ = ['AmberFormat', 'FortranFormat', 'AmberAsciiRestart', 'AmberMdcrd',
           'AmberMask', 'NetCDFTraj', 'NetCDFRestart', 'AmberOFFLibrary',
           'AmberParameterSet', 'AmberParm', 'ChamberParm', 'AmoebaParm',
           'Rst7', 'BeemanRestart', 'ConvertFromPSF', 'LoadParm', 'AMBERHOME',
           'titratable_residues']

from .amberformat import AmberFormat, FortranFormat
from .asciicrd import AmberAsciiRestart, AmberMdcrd
from .mask import AmberMask
from .netcdffiles import NetCDFTraj, NetCDFRestart
from .offlib import AmberOFFLibrary
from .parameters import AmberParameterSet
from .readparm import AmberParm, ChamberParm, AmoebaParm, Rst7, BeemanRestart, ConvertFromPSF, LoadParm
from . import titratable_residues

# See if there is an AMBERHOME defined, which we will use by default. Otherwise,
# set it to the empty string
import os as _os
AMBERHOME = _os.getenv('AMBERHOME') or ''
del _os



# These unit tests were taken from PdbxReaderTest.py and PdbxWriterTest.py from
# the pdbx package on the wwPDB website.

##
# File:    PdbxReaderTests.py
# Author:  jdw
# Date:    9-Jan-2012
# Version: 0.001
#
# Update:
#  27-Sep-2012  jdw add test case for reading PDBx structure factor file 
#
##
"""
Test cases for reading PDBx/mmCIF data files PdbxReader class -

"""
import sys, unittest, traceback, os
from utils import get_fn, get_saved_fn, diff_files, FileIOTestCase

import parmed as pmd
from parmed.exceptions import PdbxSyntaxError
from parmed.formats.pdbx import PdbxReader, PdbxWriter
from parmed.formats.pdbx.PdbxContainers import *

class PdbxReaderTests(unittest.TestCase):
    def setUp(self):
        self.verbose=False
        self.pathPdbxDataFile = get_fn("1kip.cif")
        self.pathBigPdbxDataFile = get_fn("1ffk.cif")
        self.pathSFDataFile = get_fn("1kip-sf.cif")

    def tearDown(self):
        pass

    def test_exception(self):
        """ Test the PDBx/mmCIF parser exception """
        def tmp():
            raise PdbxSyntaxError(10, 'This is the error text')
        self.assertRaises(PdbxSyntaxError, tmp)
        try:
            tmp()
        except PdbxSyntaxError as e:
            self.assertEqual(str(e), '%ERROR - [at line: 10] This is the error text')

    def test_read_small_data_file(self):
        """ Test reading small CIF file """
        try:
            #
            myDataList=[]
            ifh = open(self.pathPdbxDataFile, "r")
            pRd=PdbxReader(ifh)
            pRd.read(myDataList)
            ifh.close()            
        except:
            traceback.print_exc(file=sys.stderr)
            self.fail()

    def test_read_big_data_file(self):
        """ Test reading large CIF file """
        try:
            #
            myDataList=[]
            ifh = open(self.pathBigPdbxDataFile, "r")
            pRd=PdbxReader(ifh)
            pRd.read(myDataList)
            ifh.close()            
        except:
            traceback.print_exc(file=sys.stderr)
            self.fail()

    def test_read_sf_data_file(self):
        """read PDB structure factor data file and compute statistics on f/sig(f).
        """
        #
        myContainerList=[]
        ifh = open(self.pathSFDataFile, "r")
        pRd=PdbxReader(ifh)
        pRd.read(myContainerList)
        c0=myContainerList[0]
        #
        catObj=c0.getObj("refln")
        if catObj is None:
            return False
        
        nRows=catObj.getRowCount()
        #
        # Get column name index.
        #
        itDict={}
        itNameList=catObj.getItemNameList()
        for idxIt,itName in enumerate(itNameList):
            itDict[str(itName).lower()]=idxIt
            #
        idf=itDict['_refln.f_meas_au']
        idsigf=itDict['_refln.f_meas_sigma_au']
        minR=100
        maxR=-1
        sumR=0
        icount=0
        for row in  catObj.getRowList():
            try:
                f=float(row[idf])
                sigf=float(row[idsigf])
                ratio=sigf/f
                maxR=max(maxR,ratio)
                minR=min(minR,ratio)
                sumR+=ratio
                icount+=1
            except:
                continue
        
        ifh.close()
        self.assertAlmostEqual(minR, 0.00455182)
        self.assertAlmostEqual(maxR, 549.333333333)
        self.assertAlmostEqual(sumR/icount, 0.547058, delta=0.000002)
        self.assertEqual(icount, 18508)

    def test_space_group(self):
        parm_from_cif = pmd.load_file(get_fn('2igd.cif'))
        parm_from_pdb = pmd.load_file(get_fn('2igd.pdb'))
        self.assertEqual(parm_from_cif.space_group, 'P 21 21 21')
        self.assertEqual(parm_from_pdb.space_group, 'P 21 21 21')

class PdbxWriterTests(FileIOTestCase):
    def setUp(self):
        super(PdbxWriterTests, self).setUp()
        self.lfh=sys.stderr
        self.verbose=False
        self.pathPdbxDataFile = get_fn("1kip.cif")
        self.pathOutputFile = self.get_fn("testOutputDataFile.cif", written=True)

    def test_write_data_file(self):
        """ Test writing CIF file """
        myDataList=[]
        ofh = open(self.get_fn("test-output.cif", written=True), "w")
        curContainer=DataContainer("myblock")
        aCat=DataCategory("pdbx_seqtool_mapping_ref")
        aCat.appendAttribute("ordinal")
        aCat.appendAttribute("entity_id")
        aCat.appendAttribute("auth_mon_id")
        aCat.appendAttribute("auth_mon_num")
        aCat.appendAttribute("pdb_chain_id")
        aCat.appendAttribute("ref_mon_id")
        aCat.appendAttribute("ref_mon_num")                        
        aCat.append((1,2,3,4,5,6,7))
        aCat.append((1,2,3,4,5,6,7))
        aCat.append((1,2,3,4,5,6,7))
        aCat.append((1,2,3,4,5,6,7))
        curContainer.append(aCat)
        myDataList.append(curContainer)
        pdbxW=PdbxWriter(ofh)
        pdbxW.write(myDataList)
        ofh.close()
        self.assertTrue(
            diff_files(
                self.get_fn('test-output.cif', saved=True),
                self.get_fn('test-output.cif', written=True),
            )
        )

    def test_update_data_file(self):
        """ Test writing another CIF file """
        # Create a initial data file --
        #
        myDataList=[]
        ofh = open(self.get_fn("test-output-1.cif", written=True), "w")
        curContainer=DataContainer("myblock")
        aCat=DataCategory("pdbx_seqtool_mapping_ref")
        aCat.appendAttribute("ordinal")
        aCat.appendAttribute("entity_id")
        aCat.appendAttribute("auth_mon_id")
        aCat.appendAttribute("auth_mon_num")
        aCat.appendAttribute("pdb_chain_id")
        aCat.appendAttribute("ref_mon_id")
        aCat.appendAttribute("ref_mon_num")                        
        aCat.append((1,2,3,4,5,6,7))
        aCat.append((1,2,3,4,5,6,7))
        aCat.append((1,2,3,4,5,6,7))
        aCat.append((1,2,3,4,5,6,7))
        curContainer.append(aCat)
        myDataList.append(curContainer)
        pdbxW=PdbxWriter(ofh)
        pdbxW.write(myDataList)
        ofh.close()
        self.assertTrue(diff_files(self.get_fn('test-output-1.cif', saved=True),
                                   self.get_fn('test-output-1.cif', written=True)))
        #
        # Read and update the data -
        # 
        myDataList=[]
        ifh = open(self.get_fn("test-output-1.cif", written=True), "r")
        pRd=PdbxReader(ifh)
        pRd.read(myDataList)
        ifh.close()
        #
        myBlock=myDataList[0]
        dest = open(self.get_fn('test_write_1.txt', written=True), 'w')
        myBlock.printIt(dest)
        myCat=myBlock.getObj('pdbx_seqtool_mapping_ref')
        myCat.printIt(dest)
        dest.close()
        for iRow in range(0,myCat.getRowCount()):
            myCat.setValue('some value', 'ref_mon_id',iRow)
            myCat.setValue(100, 'ref_mon_num',iRow)
        ofh = open(self.get_fn("test-output-2.cif", written=True), "w")            
        pdbxW=PdbxWriter(ofh)
        pdbxW.write(myDataList)
        ofh.close()
        self.assertTrue(diff_files(self.get_fn('test-output-2.cif', saved=True),
                                   self.get_fn('test-output-2.cif', written=True)))
        self.assertTrue(diff_files(self.get_fn('test_write_1.txt', saved=True),
                                   self.get_fn('test_write_1.txt', written=True)))

    def test_read_data_file(self):
        """ Test reading a CIF file (... again?) """
        #
        myDataList=[]
        ifh = open(self.pathPdbxDataFile, "r")
        pRd=PdbxReader(ifh)
        pRd.read(myDataList)
        ifh.close()            

    def test_read_write_data_file(self):
        """Test case -  data file read write test
        """
        myDataList=[]
        ifh = open(self.pathPdbxDataFile, "r")            
        pRd=PdbxReader(ifh)
        pRd.read(myDataList)
        ifh.close()            

        ofh = open(self.pathOutputFile, "w")
        pWr=PdbxWriter(ofh)
        pWr.write(myDataList)        
        ofh.close()
